"""
WSGI config for HyperKitty project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/{{ docs_version }}/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application


# import sys
# import site

# For some unknown reason, sometimes mod_wsgi fails to set the python paths to
# the virtualenv, with the 'python-path' option. You can do it here too.
#
# # Remember original sys.path.
# prev_sys_path = list(sys.path)
# # Add here, for the settings module
# site.addsitedir(os.path.abspath(os.path.dirname(__file__)))
# # Add the virtualenv
# venv = os.path.join(os.path.abspath(os.path.dirname(__file__)),
#                     '..', 'lib', 'python2.6', 'site-packages')
# site.addsitedir(venv)
# # Reorder sys.path so new directories at the front.
# new_sys_path = []
# for item in list(sys.path):
#     if item not in prev_sys_path:
#         new_sys_path.append(item)
#         sys.path.remove(item)
#         sys.path[:0] = new_sys_path


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")

application = get_wsgi_application()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aamir Khan <syst3m.w0rm@gmail.com>
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
import json
import re

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import SuspiciousOperation
from django.db.models import Count
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.template import loader
from django.urls import reverse
from django.utils.timezone import utc
from django.utils.translation import gettext as _
from django.views.decorators.http import require_POST

import robot_detection
from haystack.query import SearchQuerySet

from hyperkitty.forms import AddTagForm, ReplyForm
from hyperkitty.lib.utils import stripped_subject
from hyperkitty.lib.view_helpers import (
    check_mlist_private, get_category_widget, get_months, get_posting_form)
from hyperkitty.models import Favorite, LastView, Tag, Tagging, Thread, Vote


REPLY_RE = re.compile(r'^(re:\s*)*', re.IGNORECASE)

#: GET request parameter that signifies the backend to respond with a page that
#: doesn't require Javascript to function.
NO_SCRIPT = 'noscript'


def _get_thread_replies(request, thread, limit, offset=0):
    '''
    Get and sort the replies for a thread.
    '''
    if not thread:
        raise Http404

    sort_mode = request.GET.get("sort", "thread")
    if sort_mode not in ("date", "thread"):
        raise SuspiciousOperation
    if sort_mode == "thread":
        sort_mode = "thread_order"

    mlist = request.mlist
    initial_subject = stripped_subject(mlist, thread.starting_email.subject)
    emails = list(thread
                  .emails
                  .exclude(pk=thread.starting_email.pk)
                  .select_related('sender')
                  .annotate(attachments_count=Count('attachments'))
                  .order_by(sort_mode)[offset:offset+limit])
    email_ids = [email.id for email in emails]
    votes = dict(Vote
                 .objects
                 .filter(email_id__in=email_ids, user_id=request.user.id)
                 .values_list('email_id', 'value'))
    for email in emails:
        # Extract all the votes for this message
        if request.user.is_authenticated:
            email.myvote = votes.get(email.id)
        else:
            email.myvote = None
        # Threading position
        if sort_mode == "thread_order":
            # If the email's thread_order is None, we set it to 1 by
            # default. This field should be re-computed when the async job
            # compute_thread_order_depth runs for this new email comes in
            # and fix things.
            if email.thread_order is None:
                email.thread_order = 1
            email.level = email.thread_depth - 1  # replies start ragged left
            if email.level > 5:
                email.level = 5
            elif email.level < 0:
                email.level = 0
        # Subject change
        subject = email.subject.strip()
        if mlist.subject_prefix:
            subject = subject.replace(mlist.subject_prefix, "")
        subject = REPLY_RE.sub("", subject.strip())
        if subject == initial_subject.strip():
            email.changed_subject = False
        else:
            email.changed_subject = subject
    return emails


@check_mlist_private
def thread_index(request, mlist_fqdn, threadid, month=None, year=None):
    ''' Displays all the email for a given thread identifier '''
    # The check_mlist_private decorator sets the .mlist attribute
    # on the request.
    mlist = request.mlist
    thread = get_object_or_404(Thread, mailinglist=mlist, thread_id=threadid)
    # Note(maxking): The query below fetches the thread along with it's
    # starting_email in a single complex query, instead of two rather simple
    # ones. We _could_ use the one below, but with simple tests in dev
    # environment, there isn't a major speed bump that I am seeing, so I am
    # going to keep it commented out.
    #
    # thread = Thread.objects.select_related('starting_email').get(
    #      mailinglist=mlist, thread_id=threadid)

    starting_email = thread.starting_email

    sort_mode = request.GET.get("sort", "thread")
    if request.user.is_authenticated:
        myvote = starting_email.votes.filter(user=request.user).first()
        if myvote is not None:
            starting_email.myvote = myvote.value
    else:
        starting_email.myvote = None

    # Tags
    tag_form = AddTagForm()

    # Favorites
    fav_action = "add"
    if request.user.is_authenticated and Favorite.objects.filter(
            thread=thread, user=request.user).exists():
        fav_action = "rm"

    # Category
    # category, category_form = get_category_widget(request, thread.category)

    # Extract relative dates
    today = datetime.date.today()
    days_old = today - starting_email.date.date()
    days_inactive = today - thread.date_active.date()

    subject = stripped_subject(mlist, starting_email.subject)

    # Last view
    last_view = None
    if request.user.is_authenticated:
        last_view_obj, created = LastView.objects.get_or_create(
                thread=thread, user=request.user)
        if not created:
            last_view = last_view_obj.view_date
            last_view_obj.save()  # update timestamp
    # get the number of unread messages
    if last_view is None:
        if request.user.is_authenticated:
            unread_count = thread.emails_count
        else:
            unread_count = 0
    else:
        unread_count = thread.emails.filter(date__gt=last_view).count()

    # TODO: eventually move to a middleware ?
    # http://djangosnippets.org/snippets/1865/
    user_agent = request.META.get('HTTP_USER_AGENT', None)
    if user_agent:
        is_bot = robot_detection.is_robot(user_agent)
    else:
        is_bot = True

    # Export button
    export = {
        "url": "%s?thread=%s" % (
            reverse("hk_list_export_mbox", kwargs={
                    "mlist_fqdn": mlist.name,
                    "filename": "%s-%s" % (mlist.name, thread.thread_id)}),
            thread.thread_id),
        "message": _("Download"),
        "title": _("This thread in gzipped mbox format"),
    }

    is_noscript = NO_SCRIPT in request.GET

    context = {
        'mlist': mlist,
        'thread': thread,
        'starting_email': starting_email,
        'subject': subject,
        'addtag_form': tag_form,
        'month': thread.date_active,
        'months_list': get_months(mlist),
        'days_inactive': days_inactive.days,
        'days_old': days_old.days,
        'sort_mode': sort_mode,
        'fav_action': fav_action,
        'reply_form': get_posting_form(ReplyForm, request, mlist),
        'is_bot': is_bot,
        'is_noscript': is_noscript,
        'num_comments': thread.emails_count - 1,
        'last_view': last_view,
        'unread_count': unread_count,
        'user_is_owner': mlist.is_owner(request.user),
        # XX(abraj): Commenting it out, since we don't need this. We aren't
        # really using. At some point we can remove this but I am commenting.
        #
        # 'category_form': category_form,
        # 'category': category,
        'export': export,
        'posting_enabled': getattr(
            settings, 'HYPERKITTY_ALLOW_WEB_POSTING', True),
    }

    if is_bot or is_noscript:
        # Don't rely on AJAX to load the replies
        # The limit is a safety measure, don't let a bot kill the DB
        context["replies"] = _get_thread_replies(request, thread, limit=1000)

    return render(request, "hyperkitty/thread.html", context=context)


@check_mlist_private
def replies(request, mlist_fqdn, threadid):
    """Get JSON encoded lists with the replies and the participants"""
    # chunk_size must be an even number, or the even/odd cycle will be broken.
    chunk_size = 6
    offset = int(request.GET.get("offset", "0"))
    # This is set by the @check_mlist_private decorator.
    mlist = request.mlist
    thread = get_object_or_404(Thread, mailinglist=mlist, thread_id=threadid)
    # Last view
    last_view = request.GET.get("last_view")
    if last_view:
        try:
            last_view = datetime.datetime.fromtimestamp(int(last_view), utc)
        except ValueError:
            last_view = None
    context = {
        'mlist': mlist,
        'threadid': thread,
        'reply_form': get_posting_form(ReplyForm, request, mlist),
        'last_view': last_view,
        'posting_enabled': getattr(
            settings, 'HYPERKITTY_ALLOW_WEB_POSTING', True),
    }
    context["replies"] = _get_thread_replies(request, thread, offset=offset,
                                             limit=chunk_size)

    replies_tpl = loader.get_template('hyperkitty/ajax/replies.html')
    replies_html = replies_tpl.render(context, request)
    response = {
        "replies_html": replies_html,
        "more_pending": False,
        "next_offset": None,
        }
    if len(context["replies"]) == chunk_size:
        response["more_pending"] = True
        response["next_offset"] = offset + chunk_size
    return HttpResponse(json.dumps(response),
                        content_type='application/javascript')


@require_POST
@check_mlist_private
def tags(request, mlist_fqdn, threadid):
    """ Add or remove one or more tags on a given thread. """
    if not request.user.is_authenticated:
        return HttpResponse('You must be logged in to add a tag',
                            content_type="text/plain", status=403)
    thread = get_object_or_404(
        Thread, mailinglist=request.mlist, thread_id=threadid)

    action = request.POST.get("action")

    if action == "add":
        form = AddTagForm(request.POST)
        if not form.is_valid():
            return HttpResponse("Error adding tag: invalid data",
                                content_type="text/plain", status=500)
        tagname = form.data['tag']
    elif action == "rm":
        tagname = request.POST.get('tag')
    else:
        raise SuspiciousOperation
    tagnames = [t.strip() for t in re.findall(r"[\w'_ -]+", tagname)]
    for tagname in tagnames:
        if action == "add":
            tag = Tag.objects.get_or_create(name=tagname)[0]
            Tagging.objects.get_or_create(
                tag=tag, thread=thread, user=request.user)
        elif action == "rm":
            try:
                Tagging.objects.get(tag__name=tagname, thread=thread,
                                    user=request.user).delete()
            except Tagging.DoesNotExist:
                raise Http404("No such tag: %s" % tagname)
            # cleanup
            if not Tagging.objects.filter(tag__name=tagname).exists():
                Tag.objects.filter(name=tagname).delete()

    # Now refresh the tag list
    tpl = loader.get_template('hyperkitty/threads/tags.html')
    html = tpl.render({"thread": thread}, request)

    response = {"tags": [t.name for t in thread.tags.distinct()],
                "html": html}
    return HttpResponse(json.dumps(response),
                        content_type='application/javascript')


@check_mlist_private
def suggest_tags(request, mlist_fqdn, threadid):
    term = request.GET.get("term")
    current_tags = Tag.objects.filter(
            threads__mailinglist__name=mlist_fqdn,
            threads__thread_id=threadid
        ).values_list("name", flat=True)
    if term:
        tags_db = Tag.objects.filter(
            threads__mailinglist__name=mlist_fqdn,
            name__istartswith=term)
    else:
        tags_db = Tag.objects.all()

    tag_names = list(tags_db.exclude(
        name__in=current_tags).values_list("name", flat=True)[:20])
    return HttpResponse(json.dumps(tag_names),
                        content_type='application/javascript')


@require_POST
@check_mlist_private
def favorite(request, mlist_fqdn, threadid):
    """ Add or remove from favorites"""
    if not request.user.is_authenticated:
        return HttpResponse('You must be logged in to have favorites',
                            content_type="text/plain", status=403)

    thread = get_object_or_404(
        Thread, mailinglist=request.mlist, thread_id=threadid)
    if request.POST["action"] == "add":
        Favorite.objects.get_or_create(thread=thread, user=request.user)
    elif request.POST["action"] == "rm":
        Favorite.objects.filter(thread=thread, user=request.user).delete()
    else:
        raise SuspiciousOperation
    return HttpResponse("success", content_type='text/plain')


@require_POST
@check_mlist_private
def set_category(request, mlist_fqdn, threadid):
    """ Set the category for a given thread. """
    if not request.user.is_authenticated:
        return HttpResponse('You must be logged in to add a tag',
                            content_type="text/plain", status=403)

    thread = get_object_or_404(
        Thread, mailinglist=request.mlist, thread_id=threadid)
    category, category_form = get_category_widget(request)
    if not category and thread.category:
        thread.category = None
        thread.save()
    elif category and category.name != thread.category:
        thread.category = category
        thread.save()

    # Now refresh the category widget
    context = {
            "category_form": category_form,
            "thread": thread,
            "category": category,
            }
    return render(request, "hyperkitty/threads/category.html", context)


@check_mlist_private
def reattach(request, mlist_fqdn, threadid):
    mlist = request.mlist
    if not (request.user.is_staff or
            request.user.is_superuser or
            mlist.is_owner(request.user)):
        return HttpResponse(
            'You must be a staff member or list owner to reattach a thread',
            content_type="text/plain", status=403)
    thread = get_object_or_404(Thread, mailinglist=mlist, thread_id=threadid)
    context = {
        'mlist': mlist,
        'thread': thread,
        'months_list': get_months(mlist),
    }
    template_name = "hyperkitty/reattach.html"

    if request.method == 'POST':
        parent_tid = request.POST.get("parent")
        if not parent_tid:
            parent_tid = request.POST.get("parent-manual")
        if not parent_tid or not re.match(r"\w{32}", parent_tid):
            messages.warning(
                request,
                "Invalid thread id, it should look "
                "like OUAASTM6GS4E5TEATD6R2VWMULG44NKJ.")
            return render(request, template_name, context)
        if parent_tid == threadid:
            messages.warning(
                request,
                "Can't re-attach a thread to itself, check your thread ID.")
            return render(request, template_name, context)
        try:
            parent = Thread.objects.get(
                mailinglist=thread.mailinglist, thread_id=parent_tid)
        except Thread.DoesNotExist:
            messages.warning(
                request,
                "Unknown thread, check your thread ID.")
            return render(request, template_name, context)
        try:
            thread.starting_email.set_parent(parent.starting_email)
        except ValueError as e:
            messages.error(request, str(e))
            return render(request, template_name, context)
        messages.success(request, "Thread successfully re-attached.")
        return redirect(reverse(
                'hk_thread', kwargs={
                    "mlist_fqdn": mlist_fqdn,
                    'threadid': parent_tid,
                }))
    return render(request, template_name, context)


@check_mlist_private
def reattach_suggest(request, mlist_fqdn, threadid):
    mlist = request.mlist
    thread = get_object_or_404(Thread, mailinglist=mlist, thread_id=threadid)

    default_search_query = stripped_subject(
        mlist, thread.subject).lower().replace("re:", "")
    search_query = request.GET.get("q")
    if not search_query:
        search_query = default_search_query
    search_query = search_query.strip()
    emails = SearchQuerySet().filter(
        mailinglist__exact=mlist_fqdn, content=search_query)[:50]
    suggested_threads = []
    for msg in emails:
        sugg_thread = msg.object.thread
        if (sugg_thread not in suggested_threads and
                sugg_thread.thread_id != threadid):
            suggested_threads.append(sugg_thread)

    context = {
        'mlist': mlist,
        'suggested_threads': suggested_threads[:10],
    }
    return render(request, "hyperkitty/ajax/reattach_suggest.html", context)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import logging
from urllib.error import HTTPError

from django.contrib.auth.decorators import login_required
from django.http import Http404, HttpResponse
from django.shortcuts import render
from django.urls import reverse

import dateutil.parser
import mailmanclient
from allauth.account.models import EmailAddress
from django_mailman3.lib.mailman import (
    get_mailman_client, get_mailman_user_id, get_subscriptions)
from django_mailman3.lib.paginator import paginate

from hyperkitty.lib.view_helpers import is_mlist_authorized
from hyperkitty.models import Email, Favorite, LastView, MailingList, Vote


logger = logging.getLogger(__name__)


@login_required
def user_profile(request):
    # Get the messages and paginate them
    email_addresses = EmailAddress.objects.filter(
        user=request.user).values("email")
    last_posts = Email.objects.filter(
        sender__address__in=email_addresses).order_by("-date")
    last_posts = paginate(last_posts,
                          request.GET.get("lppage"),
                          request.GET.get("lpcount", "10"))

    context = {
        "last_posts": last_posts,
        "subpage": "profile",
    }
    return render(request, "hyperkitty/user_profile/profile.html", context)


@login_required
def favorites(request):
    # Favorite threads
    favs = Favorite.objects.filter(
        user=request.user).order_by("-thread__date_active")
    favs = paginate(favs, request.GET.get('favpage'))
    return render(request, 'hyperkitty/user_profile/favorites.html', {
                "favorites": favs,
                "subpage": "favorites",
            })


@login_required
def last_views(request):
    # Last viewed threads
    lviews = LastView.objects.filter(
        user=request.user).order_by("-view_date")
    lviews = paginate(lviews, request.GET.get('lvpage'))
    return render(request, 'hyperkitty/user_profile/last_views.html', {
                "last_views": lviews,
                "subpage": "last_views",
            })


@login_required
def votes(request):
    all_votes = paginate(
        request.user.votes.order_by("-email__date").all(),
        request.GET.get('vpage'))
    return render(request, 'hyperkitty/user_profile/votes.html', {
                "votes": all_votes,
                "subpage": "votes",
            })


@login_required
def subscriptions(request):
    profile = request.user.hyperkitty_profile
    mm_user_id = get_mailman_user_id(request.user)
    subs = []
    for mlist_id in get_subscriptions(request.user):
        try:
            mlist = MailingList.objects.get(list_id=mlist_id)
        except MailingList.DoesNotExist:
            mlist = None  # no archived email yet
        posts_count = likes = dislikes = 0
        first_post = all_posts_url = None
        if mlist is not None:
            list_name = mlist.name
            posts_count = profile.emails.filter(
                mailinglist__name=mlist.name).count()
            likes, dislikes = profile.get_votes_in_list(mlist.name)
            first_post = profile.get_first_post(mlist)
            if mm_user_id is not None:
                all_posts_url = "%s?list=%s" % (
                    reverse("hk_user_posts", args=[mm_user_id]),
                    mlist.name)
        else:
            list_name = get_mailman_client().get_list(mlist_id).fqdn_listname
        likestatus = "neutral"
        if likes - dislikes >= 10:
            likestatus = "likealot"
        elif likes - dislikes > 0:
            likestatus = "like"
        subs.append({
            "list_name": list_name,
            "mlist": mlist,
            "posts_count": posts_count,
            "first_post": first_post,
            "likes": likes,
            "dislikes": dislikes,
            "likestatus": likestatus,
            "all_posts_url": all_posts_url,
        })
    return render(request, 'hyperkitty/user_profile/subscriptions.html', {
                "subscriptions": subs,
                "subpage": "subscriptions",
            })


def public_profile(request, user_id):
    class FakeMailmanUser(object):
        display_name = None
        created_on = None
        addresses = []
        subscription_list_ids = []
        user_id = None
    try:
        client = get_mailman_client()
        mm_user = client.get_user(user_id)
    except HTTPError:
        raise Http404("No user with this ID: %s" % user_id)
    except mailmanclient.MailmanConnectionError:
        mm_user = FakeMailmanUser()
        mm_user.user_id = user_id
    # XXX: don't list subscriptions, there's a privacy issue here.
    # # Subscriptions
    # subscriptions = get_subscriptions(mm_user, db_user)
    all_votes = Vote.objects.filter(email__sender__mailman_id=user_id)
    likes = all_votes.filter(value=1).count()
    dislikes = all_votes.filter(value=-1).count()
    likestatus = "neutral"
    if likes - dislikes >= 10:
        likestatus = "likealot"
    elif likes - dislikes > 0:
        likestatus = "like"
    # This is only used for the Gravatar. No email display on the public
    # profile, we have enough spam as it is, thank you very much.
    try:
        addresses = [str(addr) for addr in mm_user.addresses]
    except (KeyError, IndexError):
        addresses = []
    fullname = mm_user.display_name
    if not fullname:
        fullname = Email.objects.filter(sender__mailman_id=user_id).exclude(
                sender_name="", sender_name__isnull=True
            ).values_list("sender_name", flat=True).first()
    if mm_user.created_on is not None:
        creation = dateutil.parser.parse(mm_user.created_on)
    else:
        creation = None
    posts_count = Email.objects.filter(sender__mailman_id=user_id).count()
    is_user = request.user.is_authenticated and bool(
        set([str(a) for a in mm_user.addresses]) &
        set(request.user.hyperkitty_profile.addresses))
    context = {
        "fullname": fullname,
        "creation": creation,
        "posts_count": posts_count,
        "likes": likes,
        "dislikes": dislikes,
        "likestatus": likestatus,
        "addresses": addresses,
        "is_user": is_user,
    }
    return render(request, "hyperkitty/user_public_profile.html", context)


def posts(request, user_id):
    mlist_fqdn = request.GET.get("list")
    if mlist_fqdn is None:
        mlist = None
        return HttpResponse("Not implemented yet", status=500)
    else:
        try:
            mlist = MailingList.objects.get(name=mlist_fqdn)
        except MailingList.DoesNotExist:
            raise Http404("No archived mailing-list by that name.")
        if not is_mlist_authorized(request, mlist):
            return render(request, "hyperkitty/errors/private.html", {
                            "mlist": mlist,
                          }, status=403)

    fullname = Email.objects.filter(
            sender__mailman_id=user_id, sender_name__isnull=False
        ).exclude(sender_name="").values_list(
        "sender_name", flat=True).first()
    # Get the messages and paginate them
    emails = Email.objects.filter(
            mailinglist=mlist, sender__mailman_id=user_id
        ).order_by("date")
    emails = paginate(emails, request.GET.get("page"))

    for email in emails:
        email.myvote = email.votes.filter(user=request.user).first()

    context = {
        'user_id': user_id,
        'mlist': mlist,
        'emails': emails,
        'fullname': fullname,
    }
    return render(request, "hyperkitty/user_posts.html", context)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.

"""
Analysis of messages or threads of messages

Author: Aurelien Bompard <abompard@fedoraproject.org>
"""

from django.db import transaction

import networkx as nx


def compute_thread_order_and_depth(thread):
    # Emails must be saved, there will be DB queries in this function.
    graph = nx.DiGraph()
    thread_pos = {"d": 0, "o": 0}  # depth, order

    def walk_successors(msgid):
        obj = graph.nodes[msgid]["obj"]
        obj.thread_depth = thread_pos["d"]
        obj.thread_order = thread_pos["o"]
        obj.save(update_fields=["thread_depth", "thread_order"])
        thread_pos["d"] += 1
        thread_pos["o"] += 1
        for succ in sorted(graph.successors(msgid),
                           key=lambda m: graph.nodes[m]["num"]):
            walk_successors(succ)
        thread_pos["d"] -= 1
    for index, email in enumerate(thread.emails.order_by("date")):
        graph.add_node(email.id, num=index, obj=email)
        if email.parent_id is not None:
            graph.add_edge(email.parent_id, email.id)
            if not nx.is_directed_acyclic_graph(graph):
                # I don't want reply loops in my graph, thank you very much
                graph.remove_edge(email.parent_id, email.id)
    with transaction.atomic():
        walk_successors(thread.starting_email.id)



# -*- coding: utf-8 -*-
# flake8: noqa

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0003_thread_starting_email'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='archived_date',
            field=models.DateTimeField(default=django.utils.timezone.now, db_index=True),
        ),
        migrations.AlterField(
            model_name='email',
            name='subject',
            field=models.CharField(max_length=512, db_index=True),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#
# XXX(maxking): Category system is entirely un-used in Hyperkitty currently, we
# should figure out what do we want it to look like and then re-enable it at
# some point.

import re

from django.core.validators import RegexValidator
from django.db import models
from django.forms import TextInput


# Max length of a color's hex code is 7, which includes a preceding '#'.
MAX_COLOR_LENGTH = 7
# Regex to validate colors which are HEX of length 6, preceded by a '#' sign.
color_re = re.compile('^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$')
# Custom validator for HEX color.
validate_color = RegexValidator(
    regex=color_re,
    # Error message used if the validation fails.
    message='Enter a valid hex color.',
    # Error code used if validation fails.
    code='invalid')


class ColorInput(TextInput):
    """Form widget to use for a color input."""
    input_type = 'color'


class ColorField(models.CharField):

    default_validators = [validate_color]

    def __init__(self, *args, **kwargs):
        kwargs['max_length'] = MAX_COLOR_LENGTH
        super(ColorField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs['widget'] = ColorInput
        return super(ColorField, self).formfield(**kwargs)


class ThreadCategory(models.Model):
    name = models.CharField(max_length=255, db_index=True, unique=True)
    color = ColorField()

    class Meta:
        verbose_name_plural = "Thread categories"

    def __str__(self):
        return 'Thread category "%s"' % self.name



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.

"""
Definition of async tasks using Django-Q.

Author: Aurelien Bompard <abompard@fedoraproject.org>
"""

import logging

from django.core.cache import cache
from django.core.cache.utils import make_template_fragment_key

from django_q.tasks import AsyncTask
from mailmanclient import MailmanConnectionError

from hyperkitty.lib.analysis import compute_thread_order_and_depth
from hyperkitty.lib.utils import run_with_lock
from hyperkitty.models.email import Email
from hyperkitty.models.mailinglist import MailingList
from hyperkitty.models.sender import Sender
from hyperkitty.models.thread import Thread
from hyperkitty.search_indexes import update_index


log = logging.getLogger(__name__)

#
# Tasks
#


def update_search_index():
    """Update the full-text index"""
    AsyncTask(run_with_lock, update_index, remove=False,
              q_options={'task_name': 'update_search_index'}).run()


# def update_and_clean_search_index():
#     """Update the full-text index and clean old entries"""
#     run_with_lock(update_index, remove=True)

def rebuild_mailinglist_cache_recent(mlist_name):
    AsyncTask(
        _rebuild_mailinglist_cache_recent, mlist_name,
        q_options={'task_name': 'rebuild_mailinglist_cache_recent'}).run()


def _rebuild_mailinglist_cache_recent(mlist_name):
    try:
        mlist = MailingList.objects.get(name=mlist_name)
    except MailingList.DoesNotExist:
        log.warning('MailingList with name %s does not exist', mlist_name)
        return
    for cached_value in mlist.recent_cached_values:
        cached_value.rebuild()


def rebuild_mailinglist_cache_for_month(mlist_name, year, month):
    AsyncTask(
        _rebuild_mailinglist_cache_for_month, mlist_name, year, month,
        q_options={'task_name': 'rebuild_mailinglist_cache_for_month'}).run()


def _rebuild_mailinglist_cache_for_month(mlist_name, year, month):
    try:
        mlist = MailingList.objects.get(name=mlist_name)
    except MailingList.DoesNotExist:
        log.warning('MailingList with name %s does not exist', mlist_name)
        return
    mlist.cached_values["participants_count_for_month"].rebuild(year, month)


def rebuild_thread_cache_new_email(thread_id):
    AsyncTask(_rebuild_thread_cache_new_email, thread_id,
              q_options={'task_name': 'rebuild_thread_cache_new_email'}).run()


def _rebuild_thread_cache_new_email(thread_id):
    try:
        thread = Thread.objects.get(id=thread_id)
    except Thread.DoesNotExist:
        log.warning(
            "Cannot rebuild the thread cache: thread %s does not exist.",
            thread_id)
        return
    for cached_key in ["participants_count", "emails_count"]:
        thread.cached_values[cached_key].rebuild()
    # Don't forget the cached template fragment.
    cache.delete(make_template_fragment_key(
        "thread_participants", [thread.id]))


def rebuild_cache_popular_threads(mlist_name):
    AsyncTask(_rebuild_cache_popular_threads, mlist_name,
              q_options={'task_name': 'rebuild_cache_popular_threads'}).run()


def _rebuild_cache_popular_threads(mlist_name):
    try:
        mlist = MailingList.objects.get(name=mlist_name)
    except MailingList.DoesNotExist:
        log.warning('MailingList with name %s does not exist', mlist_name)
        return
    mlist.cached_values["popular_threads"].rebuild()


def compute_thread_positions(thread_id):
    AsyncTask(_compute_thread_positions, thread_id,
              q_options={'task_name': 'compute_thread_positions'}).run()


def _compute_thread_positions(thread_id):
    try:
        thread = Thread.objects.get(id=thread_id)
    except Thread.DoesNotExist:
        # Maybe the thread was deleted? Not much we can do here.
        log.warning(
            "Cannot rebuild the thread cache: thread %s does not exist.",
            thread_id)
        return
    compute_thread_order_and_depth(thread)


def update_from_mailman(mlist_name):
    AsyncTask(_update_from_mailman, mlist_name,
              q_options={'task_name': 'update_from_mailman'}).run()


def _update_from_mailman(mlist_name):
    try:
        mlist = MailingList.objects.get(name=mlist_name)
    except MailingList.DoesNotExist:
        log.warning('MailingList with name %s does not exist', mlist_name)
        return
    mlist.update_from_mailman()


def sender_mailman_id(sender_id):
    AsyncTask(_sender_mailman_id, sender_id,
              q_options={'task_name': 'sender_mailman_id'}).run()


def _sender_mailman_id(sender_id):
    sender = Sender.objects.get(pk=sender_id)
    try:
        sender.set_mailman_id()
    except MailmanConnectionError:
        pass


def check_orphans(email_id):
    AsyncTask(_check_orphans, email_id,
              q_options={'task_name': 'check_orphans'}).run()


def _check_orphans(email_id):
    """
    When a reply is received before its original message, it must be
    re-attached when the original message arrives.
    """
    try:
        email = Email.objects.get(id=email_id)
    except Email.DoesNotExist:
        # Maybe the email was deleted? Not much we can do here.
        log.warning(
            "Cannot check for orphans: email %s does not exist.", email_id)
        return
    orphans = Email.objects.filter(
            mailinglist=email.mailinglist,
            in_reply_to=email.message_id,
            parent_id__isnull=True,
        ).exclude(
            # guard against emails with the in-reply-to header pointing to
            # themselves
            id=email.id
        )
    for orphan in orphans:
        orphan.set_parent(email)


def rebuild_thread_cache_votes(thread_id):
    AsyncTask(_rebuild_thread_cache_votes, thread_id,
              q_options={'task_name': 'rebuild_thread_cache_votes'}).run()


def _rebuild_thread_cache_votes(thread_id):
    try:
        thread = Thread.objects.get(id=thread_id)
    except Thread.DoesNotExist:
        log.warning('Thread with id {thread_id} does not exist'.format(
            thread_id=thread_id))
        return

    for cached_key in ["votes", "votes_total"]:
        thread.cached_values[cached_key].rebuild()


def rebuild_email_cache_votes(email_id):
    AsyncTask(_rebuild_email_cache_votes, email_id,
              q_options={'task_name': 'rebuild_email_cache_votes'}).run()


def _rebuild_email_cache_votes(email_id):
    email = Email.objects.get(id=email_id)
    email.cached_values["votes"].rebuild()



# -*- coding: utf-8 -*-

from django.db import connection, migrations


def update_thread_and_email(apps, schema_editor):
    """
    Update Thread & Email tables's mailinglist_id column to the new
    MailingList.id value.
    """
    # import sys
    # print(" Updating thread and email references. This will take a "
    #       "loooooong time, go get a coffee. ", end="")
    # sys.stdout.flush()
    # # This it the version using the models. It is slow.
    # MailingList = apps.get_model("hyperkitty", "MailingList")
    # Thread = apps.get_model("hyperkitty", "Thread")
    # Email = apps.get_model("hyperkitty", "Email")
    # mlists = dict([(ml.name, ml.id) for ml in MailingList.objects.all()])
    # for thread in Thread.objects.all():
    #     thread.mailinglist = mlists[thread.mailinglist]
    #     thread.save()
    # for email in Email.objects.all():
    #     email.mailinglist = mlists[email.mailinglist]
    #     email.save()
    # This is the version using only two queries. It is much faster.
    with connection.cursor() as cursor:
        cursor.execute("""
            UPDATE hyperkitty_thread SET mailinglist_id = (
                SELECT id FROM hyperkitty_mailinglist
                WHERE name = hyperkitty_thread.mailinglist_id LIMIT 1
            )
            """)
        cursor.execute("""
            UPDATE hyperkitty_email SET mailinglist_id = (
                SELECT id FROM hyperkitty_mailinglist
                WHERE name = hyperkitty_email.mailinglist_id LIMIT 1
            )
            """)


def update_thread_and_email_reverse(apps, schema_editor):
    with connection.cursor() as cursor:
        cursor.execute("""
            UPDATE hyperkitty_thread SET mailinglist_id = (
                SELECT name FROM hyperkitty_mailinglist
                WHERE id = hyperkitty_thread.mailinglist_id LIMIT 1
            )
            """)
        cursor.execute("""
            UPDATE hyperkitty_email SET mailinglist_id = (
                SELECT name FROM hyperkitty_mailinglist
                WHERE id = hyperkitty_email.mailinglist_id LIMIT 1
            )
            """)


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0013_mailinglist_id_1'),
    ]

    operations = [
        migrations.RunPython(update_thread_and_email,
                             update_thread_and_email_reverse),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from xml.sax.saxutils import escape

from django import template


register = template.Library()


@register.filter
def nolongterms(text):
    """
    Remove terms longer than 245 chars, or Xapian will choke on them:
    https://github.com/notanumber/xapian-haystack/issues/77
    The size check is done on the XML-escaped and quotes-escaped encoded
    version of the term, with the XTEXT prefix (thus the '240' limit)
    """
    def _getlen(word):
        return len(
            escape(word).replace(
                '"', '&quot;').replace(
                "'", "&#39;").encode("utf-8"))
    if not text:
        return ""
    return ' '.join(word for word in text.split() if _getlen(word) < 240)



# Generated by Django 2.0.9 on 2019-01-27 04:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0018_threadcategory_color'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mailinglist',
            name='description',
            field=models.TextField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import os
import random

from django.conf import settings
from django.core.checks import Error, register


@register()
def config_check(app_configs, **kwargs):
    if app_configs is not None and 'hyperkitty' not in [
            c.name for c in app_configs]:
        return []
    attachment_folder = getattr(
        settings, "HYPERKITTY_ATTACHMENT_FOLDER", None)
    errors = []
    if attachment_folder is not None:
        if not os.path.exists(attachment_folder):
            errors.append(
                Error(
                    'HyperKitty\'s attachment folder does not exist',
                    hint=('The folder set in the config variable '
                          'HYPERKITTY_ATTACHMENT_FOLDER does not exist yet '
                          '({!r}). You must create it and make sure the '
                          'webserver has the permissions to write there.'
                          ).format(attachment_folder),
                    id='hyperkitty.E001',
                )
            )
        else:
            extra = random.randrange(10000)
            try:
                filepath = os.path.join(attachment_folder, f'_check_{extra}')
                with open(filepath, "w") as f:
                    f.write("check")
                os.remove(filepath)
            except OSError:
                errors.append(
                    Error(
                        'Could not write to HyperKitty\'s attachment folder',
                        hint='The folder set in the config variable '
                             'HYPERKITTY_ATTACHMENT_FOLDER cannot be '
                             'written to. Make sure the webserver has '
                             'the permissions to write there.',
                        id='hyperkitty.E002',
                    )
                )
    return errors



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>


import email.utils
import logging
import os
import os.path
import re
from base64 import b32encode
from contextlib import contextmanager
from datetime import timedelta
from email.parser import BytesHeaderParser, HeaderParser
from email.policy import default
from hashlib import sha1
from tempfile import gettempdir

from django.conf import settings
from django.db import connection
from django.utils import timezone

import dateutil.parser
import dateutil.tz
from flufl.lock import Lock


log = logging.getLogger(__name__)


def get_message_id_hash(msg_id):
    """
    Returns the X-Message-ID-Hash header for the provided Message-ID header.

    See <http://wiki.list.org/display/DEV/Stable+URLs#StableURLs-Headers> for
    details. Example:
    """
    msg_id = email.utils.unquote(msg_id).encode('utf-8')
    return b32encode(sha1(msg_id).digest()).decode('utf-8')


def get_message_id(message):
    msg_id = email.utils.unquote(re.sub(r'\s', '', message['Message-Id']))
    # Protect against extremely long Message-Ids (there is no limit in the
    # email spec), it's set to VARCHAR(255) in the database
    if len(msg_id) >= 255:
        msg_id = msg_id[:254]
    return msg_id


IN_BRACKETS_RE = re.compile("[^<]*<([^>]+)>.*")


def get_ref(message):
    """
    Returns the message-id of the reference email for a given message.
    """
    if ("References" not in message and
            "In-Reply-To" not in message):
        return None
    ref_id = message.get("In-Reply-To")

    # EmailMessage will always return instances of str
    assert ref_id is None or isinstance(ref_id, str)

    if ref_id is None or not ref_id.strip():
        ref_id = message.get("References")
        if ref_id is not None and ref_id.strip():
            # There can be multiple references, use the last one
            ref_id = ref_id.split()[-1].strip()
    if ref_id is not None:
        if "<" in ref_id or ">" in ref_id:
            ref_id = IN_BRACKETS_RE.match(ref_id)
            if ref_id:
                ref_id = ref_id.group(1)
    if ref_id is not None:
        ref_id = ref_id[:254]
    return ref_id


def parseaddr(address):
    """
    Wrapper around email.utils.parseaddr to also handle Mailman's generated
    mbox archives.
    """
    if address is None:
        return "", ""
    from_name, from_email = email.utils.parseaddr(address)
    if '@' not in from_email:
        address = address.replace(" at ", "@")
        from_name, from_email = email.utils.parseaddr(address)
    if not from_name:
        from_name = from_email
    return from_name, from_email


def parsedate(datestring):
    if datestring is None:
        return None
    try:
        parsed = dateutil.parser.parse(datestring)
    except ValueError:
        return None
    try:
        offset = parsed.utcoffset()
    except ValueError:
        # Wrong offset, reset to UTC
        offset = None
        parsed = parsed.replace(tzinfo=timezone.utc)
    if offset is not None and \
            abs(offset) > timedelta(hours=13):
        parsed = parsed.astimezone(timezone.utc)
    if parsed.tzinfo is None:
        parsed = parsed.replace(tzinfo=timezone.utc)  # make it aware
    return parsed


def header_to_unicode(header):
    if header is None:
        header = str(header)
    if isinstance(header, str):
        msg = HeaderParser(policy=default).parsestr('dummy: ' + header)
    elif isinstance(header, bytes):
        msg = BytesHeaderParser(policy=default).parsebytes(b'dummy: ' + header)
    else:
        raise ValueError('header must be str or bytes, but is ' + type(header))

    return msg['dummy']


def stripped_subject(mlist, subject):
    if mlist is None:
        return subject
    if not subject:
        return "(no subject)"
    if not mlist.subject_prefix:
        return subject
    if subject.lower().startswith(mlist.subject_prefix.lower()):
        subject = subject[len(mlist.subject_prefix):]
    return subject


# File-based locking
def run_with_lock(fn, *args, **kwargs):
    if kwargs.get('remove'):
        # remove = True is slow. We need to extend the lock life
        lock_life = getattr(settings,
                            "HYPERKITTY_JOBS_UPDATE_INDEX_LOCK_LIFE", 900)
    else:
        # Use the default (15 sec)
        lock_life = None
    lock = Lock(getattr(
        settings, "HYPERKITTY_JOBS_UPDATE_INDEX_LOCKFILE",
        os.path.join(gettempdir(), "hyperkitty-jobs-update-index.lock")),
        lifetime=lock_life)
    if lock.is_locked:
        log.warning(
            "Update index lock is acquired by: {}".format(*lock.details))
        return
    with lock:
        try:
            fn(*args, **kwargs)
        except Exception as e:
            log.exception("Failed to update the fulltext index: %s", e)


@contextmanager
def pgsql_disable_indexscan():
    # Sometimes PostgreSQL chooses a very inefficient query plan:
    # https://pagure.io/fedora-infrastructure/issue/6164
    if connection.vendor != "postgresql":
        yield
        return
    with connection.cursor() as cursor:
        cursor.execute("SET enable_indexscan = OFF")
        try:
            yield
        finally:
            cursor.execute("SET enable_indexscan = ON")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
import json
import zlib

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist
from django.db import IntegrityError, transaction
from django.http import (
    Http404, HttpResponse, HttpResponseBadRequest, StreamingHttpResponse)
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import formats, timezone
from django.utils.dateformat import format as date_format
from django.utils.translation import gettext as _
from django.views.decorators.cache import cache_page

from django_mailman3.lib.mailman import get_mailman_user_id
from django_mailman3.lib.paginator import paginate

from hyperkitty.lib.view_helpers import (
    check_mlist_private, daterange, get_display_dates, get_months)
from hyperkitty.models import Favorite, LastView
from hyperkitty.signals import silenced_email_pre_delete


@check_mlist_private
def archives(request, mlist_fqdn, year=None, month=None, day=None):
    """List of threads in MailingList.

    If year & month is None, we return *all* the threads and render a view with
    all the threads of a MailingList.
    """
    mlist = request.mlist

    if year is None and month is None:
        # If the year and month is None, we set the begin date to the date of
        # the first email and end date to be today.
        end_date = datetime.date.today()
        # Since we don't need any special filtering, just return *all* the
        # threads ordered by date_active.
        threads = (mlist
                   .threads
                   .order_by("-date_active")
                   .select_related('starting_email')
                   .select_related('starting_email__sender'))

        # Get the earliest email in the mailing list to get the begin_date for
        # the mailing list.
        try:
            begin_date = (mlist
                          .emails
                          .order_by('date')
                          .values_list('date', flat=True)[0])
        except IndexError:
            # If there aren't any emails for the list, we just put the current
            # date as the begin and end date both.
            begin_date = end_date

        # Set the month and year to be today's.
        year = end_date.year
        month = end_date.month
        # The list title for all the threads.
        list_title = ""
        no_results_text = _("for this MailingList")
    else:
        try:
            begin_date, end_date = get_display_dates(year, month, day)
        except ValueError:
            # Wrong date format, for example 9999/0/0
            raise Http404("Wrong date format")

        threads = mlist.get_threads_between(begin_date, end_date)

        if day is None:
            list_title = date_format(begin_date, "F Y")
            no_results_text = _("for this month")
        else:
            list_title = formats.date_format(begin_date)  # works with i18n
            no_results_text = _("for this day")

    # Export button
    export = {
        "url": "%s?start=%s&end=%s" % (
            reverse("hk_list_export_mbox", kwargs={
                    "mlist_fqdn": mlist.name,
                    "filename": "%s-%s" % (
                        mlist.name, end_date.strftime("%Y-%m"))}),
            begin_date.strftime("%Y-%m-%d"),
            end_date.strftime("%Y-%m-%d")),
        "message": _("Download"),
        "title": _("This month in gzipped mbox format"),
    }
    extra_context = {
        'month': begin_date,
        'month_num': begin_date.month,
        "list_title": list_title.capitalize(),
        "no_results_text": no_results_text,
        "export": export,
    }
    if day is None:
        extra_context["participants_count"] = \
            mlist.get_participants_count_for_month(int(year), int(month))
    return _thread_list(request, mlist, threads, extra_context=extra_context)


def _thread_list(request, mlist, threads,
                 template_name='hyperkitty/thread_list.html',
                 extra_context=None):
    threads = paginate(threads,
                       request.GET.get('page'),
                       request.GET.get('count'))

    # We need to set favorites only if the user is logged in.
    if request.user.is_authenticated:
        # First, make a list of all the threads.
        thread_ids = [thread.id for thread in threads]
        # Then, get all the rows in the favs table which correspond to the
        # current user but the thread ids is from the above list. This will
        # get us which of the above threads are favorited, on which we should
        # set the favorite to be true.
        favs = (Favorite
                .objects
                .filter(thread_id__in=thread_ids, user=request.user)
                .values_list('thread_id', flat=True))

        # Do the same dance for un-read thing of the user.

        # This is a re-implementation of Thread.is_unread_by method that gets
        # multiple values in a single query and does the actual computation in
        # memory for unread count.
        last_views = (LastView
                      .objects
                      .filter(thread_id__in=thread_ids, user=request.user)
                      .values_list('id', 'thread_id', 'view_date'))

        # If there are more than one entry for single thread, delete the older
        # one and keep only the new one.
        lvs = {}
        for view_id, thread_id, view_date in last_views:
            # If we've seen this thread_id, i.e. this user has duplicate
            # LastView entries for this thread, so just delete the older one.
            if thread_id in lvs:
                if lvs[thread_id][1] > view_date:
                    del_id = view_id
                    lvs[thread_id] = (view_id, view_date)
                else:
                    # The one in lvs is the older one, so we'll delete that.
                    del_id = lvs[thread_id][0]
                LastView.objects.delete(pk=del_id)
            else:
                lvs[thread_id] = (view_id, view_date)

        for thread in threads:
            # Set if the user favorited this thread.
            thread.favorite = thread.id in favs
            # Sef if the user read this thread based on LastViews.
            if thread.id not in lvs:
                thread.is_unread = True
            else:
                thread.is_unread = (
                    thread.date_active.replace(tzinfo=timezone.utc) > lvs.get(thread.id)[1])  # noqa: E501

    context = {
        'mlist': mlist,
        'threads': threads,
        'months_list': get_months(mlist),
    }
    if extra_context is not None:
        context.update(extra_context)
    return render(request, template_name, context)


@check_mlist_private
def overview(request, mlist_fqdn=None):
    if not mlist_fqdn:
        return redirect('/')
    mlist = request.mlist

    # top authors are the ones that have the most kudos.  How do we determine
    # that?  Most likes for their post?
    authors = []

    # Threads by category
    threads_by_category = {}

    # Export button
    recent_dates = [
        d.strftime("%Y-%m-%d") for d in mlist.get_recent_dates()]
    recent_url = "%s?start=%s&end=%s" % (
        reverse("hk_list_export_mbox", kwargs={
                "mlist_fqdn": mlist.name,
                "filename": "%s-%s-%s" % (
                    mlist.name, recent_dates[0], recent_dates[1])}),
        recent_dates[0], recent_dates[1])
    today = datetime.date.today()
    month_dates = get_display_dates(today.year, today.month, None)
    month_url = "%s?start=%s&end=%s" % (
        reverse("hk_list_export_mbox", kwargs={
                "mlist_fqdn": mlist.name,
                "filename": "%s-%s" % (mlist.name, today.strftime("%Y-%m"))}),
        month_dates[0].strftime("%Y-%m-%d"),
        month_dates[1].strftime("%Y-%m-%d"))
    export = {"recent": recent_url, "month": month_url}

    context = {
        'view_name': 'overview',
        'mlist': mlist,
        'top_author': authors,
        'threads_by_category': threads_by_category,
        'months_list': get_months(mlist),
        'export': export,
        'posting_enabled': getattr(
            settings, 'HYPERKITTY_ALLOW_WEB_POSTING', True),
    }
    return render(request, "hyperkitty/overview.html", context)


@check_mlist_private
# @cache_page(3600 * 12)  # cache for 12 hours
def overview_recent_threads(request, mlist_fqdn):
    """Return the most recently updated threads."""
    mlist = request.mlist
    return render(request, "hyperkitty/fragments/overview_threads.html", {
        'mlist': mlist,
        'threads': mlist.recent_threads[:20],
        'empty': _('No discussions this month (yet).'),
        })


@check_mlist_private
# @cache_page(3600 * 12)  # cache for 12 hours
def overview_pop_threads(request, mlist_fqdn):
    """Return the threads with the most votes."""
    mlist = request.mlist
    return render(request, "hyperkitty/fragments/overview_threads.html", {
        'mlist': mlist,
        'threads': mlist.popular_threads,
        "empty": _('No vote has been cast this month (yet).'),
        })


@check_mlist_private
# @cache_page(3600 * 12)  # cache for 12 hours
def overview_top_threads(request, mlist_fqdn):
    """Return the threads with the most answers."""
    mlist = request.mlist
    return render(request, "hyperkitty/fragments/overview_threads.html", {
        'mlist': mlist,
        'threads': mlist.top_threads,
        "empty": _('No discussions this month (yet).'),
        })


@check_mlist_private
# @cache_page(3600 * 12)  # cache for 12 hours
def overview_favorites(request, mlist_fqdn):
    """Return the threads that the logged-in user has set as favorite."""
    mlist = request.mlist
    if request.user.is_authenticated:
        favorites = [f.thread for f in Favorite.objects.filter(
            thread__mailinglist=mlist, user=request.user)]
    else:
        favorites = []
    return render(request, "hyperkitty/fragments/overview_threads.html", {
        'mlist': mlist,
        'threads': favorites,
        "empty": _('You have not flagged any discussions (yet).'),
        })


@check_mlist_private
# @cache_page(3600 * 12)  # cache for 12 hours
def overview_posted_to(request, mlist_fqdn):
    """Return the threads that the logged-in user has posted to."""
    mlist = request.mlist
    if request.user.is_authenticated:
        mm_user_id = get_mailman_user_id(request.user)
        threads_posted_to = []
        if mm_user_id is not None:
            for thread in mlist.recent_threads:
                senders = set(
                    [e.sender.mailman_id for e in thread.emails.all()])
                if mm_user_id in senders:
                    threads_posted_to.append(thread)
    else:
        threads_posted_to = []
    return render(request, "hyperkitty/fragments/overview_threads.html", {
        'mlist': mlist,
        'threads': threads_posted_to,
        "empty": _('You have not posted to this list (yet).'),
        })


@check_mlist_private
# @cache_page(3600 * 12)  # cache for 12 hours
def overview_top_posters(request, mlist_fqdn):
    """Return the authors that sent the most emails."""
    mlist = request.mlist
    return render(request, "hyperkitty/fragments/overview_top_posters.html", {
        'mlist': mlist,
        })


@check_mlist_private
@cache_page(3600 * 12)  # cache for 12 hours
def recent_activity(request, mlist_fqdn):
    """Return the number of emails posted in the last 30 days"""
    mlist = request.mlist
    begin_date, end_date = mlist.get_recent_dates()
    days = daterange(begin_date, end_date)

    # Use get_messages and not get_threads to count the emails, because
    # recently active threads include messages from before the start date
    emails_in_month = mlist.emails.filter(
        date__gte=begin_date,
        date__lt=end_date)
    # graph
    emails_per_date = {}
    # populate with all days before adding data.
    for day in days:
        emails_per_date[day.strftime("%Y-%m-%d")] = 0
    # now count the emails
    for email in emails_in_month:
        date_str = email.date.strftime("%Y-%m-%d")
        if date_str not in emails_per_date:
            continue  # outside the range
        emails_per_date[date_str] += 1
    # return the proper format for the javascript chart function
    evolution = [{"date": d, "count": emails_per_date[d]}
                 for d in sorted(emails_per_date)]
    return HttpResponse(json.dumps({"evolution": evolution}),
                        content_type='application/javascript')


@check_mlist_private
def export_mbox(request, mlist_fqdn, filename):
    # If the settings HYPERKITTY_MBOX_EXPORT is set to False, then
    # disable download of archives.
    if not getattr(settings, "HYPERKITTY_MBOX_EXPORT", True):
        return HttpResponseBadRequest("Archive download disabled.")

    mlist = request.mlist
    query = mlist.emails
    try:
        if "start" in request.GET:
            start_date = datetime.datetime.strptime(
                request.GET["start"], "%Y-%m-%d")
            start_date = timezone.make_aware(start_date, timezone.utc)
            query = query.filter(date__gte=start_date)
        if "end" in request.GET:
            end_date = datetime.datetime.strptime(
                request.GET["end"], "%Y-%m-%d")
            end_date = timezone.make_aware(end_date, timezone.utc)
            query = query.filter(date__lt=end_date)
    except ValueError:
        return HttpResponseBadRequest("Invalid dates")
    if "thread" in request.GET:
        query = query.filter(thread__thread_id=request.GET["thread"])
    if "message" in request.GET:
        query = query.filter(message_id_hash=request.GET["message"])

    def stream_mbox(query):
        # Use the gzip format: http://www.zlib.net/manual.html#Advanced
        compressor = zlib.compressobj(6, zlib.DEFLATED, zlib.MAX_WBITS | 16)
        for email in query.order_by("archived_date").all():
            yield compressor.compress(email.as_bytes())
            yield compressor.compress(b"\n\n")
        yield compressor.flush()
    response = StreamingHttpResponse(
        stream_mbox(query), content_type="application/gzip")
    response['Content-Disposition'] = (
        'attachment; filename="%s.mbox.gz"' % filename)
    return response


@login_required
@check_mlist_private
@transaction.atomic
def delete(request, mlist_fqdn):
    mlist = request.mlist
    if not request.user.is_staff and not request.user.is_superuser:
        return HttpResponse(
            _('You must be a staff member to delete a MailingList'),
            content_type="text/plain", status=403)
    if request.method == 'POST':
        # XXX(maxking): This is a hack to make sure that we don't continuously
        # keep re-setting the parent of emails when trying to delete all the
        # emails and threads.
        # We silence the pre_delete hook for email which does this re-balancing
        # and then continue to do the delete action.
        with silenced_email_pre_delete():
            try:
                mlist.threads.all().delete()
                mlist.delete()
                messages.success(
                    request,
                    _('Successfully deleted {}'.format(mlist.name)))
                return redirect('/')
            except (IntegrityError, ObjectDoesNotExist) as e:
                messages.error(request, str(e))

            return redirect(
                reverse('hk_list_overview', kwargs={"mlist_fqdn": mlist.name}))
    else:
        return render(request, "hyperkitty/list_delete.html", {"mlist": mlist})



# Generated by Django 3.2.11 on 2022-02-13 01:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0022_mailinglist_archive_rendering_mode'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mailinglist',
            name='name',
            field=models.CharField(db_index=True, max_length=254, unique=True),
        ),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa

from django.db import IntegrityError, connection, migrations, models, utils


PROVIDERS_MAP = {
    "fedora": "fedora",
    "yahoo": "openid",
}


def populate_emailaddress(apps, schema_editor):
    # All current users have verified their email. Populate the EmailAddress
    # and mark is as such.
    User = apps.get_model("auth", "User")
    EmailAddress = apps.get_model("account", "EmailAddress")
    for user in User.objects.all():
        if not EmailAddress.objects.filter(email=user.email).exists():
            EmailAddress.objects.create(
                email=user.email, user=user,
                verified=True, primary=True,
                )


def migrate_social_users(apps, schema_editor):
    # Migrate the Social Auth association to AllAuth
    SocialAccount = apps.get_model("socialaccount", "SocialAccount")
    # We can't use the UserSocialAuth model because the social_auth app has
    # been removed, and thus the model isn't available.
    cursor = connection.cursor()
    if connection.vendor == "postgresql":
        # a select on a non-existing table in pg, marks the tx abort only.
        cursor.execute("select 1 from information_schema.tables where table_name = 'social_auth_usersocialauth' and  to_regclass('social_auth_usersocialauth') is not null")
        if not cursor.fetchall():
          # No social_auth table, stop here.
          return
    else:
        try:
            cursor.execute("SELECT 1 from social_auth_usersocialauth")
        except (utils.OperationalError, utils.ProgrammingError):
            # No social_auth table, stop here.
            return

    for provider_old, provider_new in PROVIDERS_MAP.items():
        cursor.execute("""
            SELECT uid, user_id, last_login, date_joined
            FROM social_auth_usersocialauth usa
            JOIN auth_user ON usa.user_id = auth_user.id
            WHERE provider = %s
            """, (provider_old,))
        for row in cursor:
            uid, user_id, last_login, date_joined = row
            if not SocialAccount.objects.filter(
                    provider=provider_new, uid=uid).exists():
                SocialAccount.objects.create(
                        provider=provider_new,
                        uid=uid,
                        user_id=user_id,
                        last_login=last_login,
                        date_joined=date_joined,
                        extra_data={},
                    )


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0006_thread_on_delete'),
        ('socialaccount', '0003_extra_data_default_dict'),
        ('account', '0002_email_max_length'),
    ]

    operations = [
        migrations.RunPython(populate_emailaddress),
        migrations.RunPython(migrate_social_users),
    ]



# -*- coding: utf-8 -*-
# flake8: noqa
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='in_reply_to',
            field=models.CharField(db_index=True, max_length=255, null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
import re

from django.http import Http404

from hyperkitty.models import MailingList


PORT_IN_URL = re.compile(r':\d+$')


def get_list_by_name(list_name, domain):
    matching = list(MailingList.objects.filter(name__startswith=list_name+"@"))
    if len(matching) == 0:  # no candidate found
        raise Http404("No archived mailinglist by that name")
    if len(matching) == 1:  # only one candidate
        return matching[0]

    # more than one result, try using the hostname
    domain = PORT_IN_URL.sub('', domain)
    list_fqdn = "%s@%s" % (list_name, domain)
    try:
        return MailingList.objects.get(name=list_fqdn)
    except MailingList.DoesNotExist:
        # return the first match, arbitrarily
        return matching[0]


def month_name_to_num(month_name):
    """map month names to months numbers"""
    months = dict((datetime.date(2000, num, 1).strftime('%B'), num)
                  for num in range(1, 12))
    return months[month_name]



# -*- coding: utf-8 -*-
# Copyright (C) 1998-2017 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.

"""
This file is the main URL config for a Django website including HyperKitty.
"""

from django.conf import settings
from django.conf.urls import include
from django.contrib import admin
from django.urls import path, reverse_lazy
from django.views.generic import RedirectView


urlpatterns = [
    path('', RedirectView.as_view(
        url=reverse_lazy('hk_root'))),
    path('hyperkitty/', include('hyperkitty.urls')),
    # url(r'^postorius/', include('postorius.urls')),
    path('', include('django_mailman3.urls')),
    path('accounts/', include('allauth.urls')),
    # Django admin
    path('admin/', admin.site.urls),
]


if settings.DEBUG:
    import debug_toolbar
    urlpatterns = [
        path('__debug__/', include(debug_toolbar.urls)),
    ] + urlpatterns



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import logging
import re
from email.message import EmailMessage

from django.conf import settings
from django.db import DataError
from django.utils import timezone

from django_mailman3.lib.scrub import Scrubber

from hyperkitty.lib.utils import (
    get_message_id, get_ref, header_to_unicode, parseaddr, parsedate)
from hyperkitty.models import (
    ArchivePolicy, Attachment, Email, MailingList, Sender)
from hyperkitty.tasks import sender_mailman_id, update_from_mailman


logger = logging.getLogger(__name__)


UNIXFROM_DATE_RE = re.compile(r'^\s*[^\s]+@[^\s]+ (.*)$')


class DuplicateMessage(Exception):
    """
    The database already contains an email with the same Message-ID header.
    """


def add_to_list(list_name, message, from_import=False):
    assert isinstance(message, EmailMessage)
    # timeit("1 start")
    mlist = MailingList.objects.get_or_create(name=list_name)[0]
    if not getattr(settings, "HYPERKITTY_BATCH_MODE", False):
        update_from_mailman(mlist.name)
    mlist.save()
    if mlist.archive_policy == ArchivePolicy.never.value and not from_import:
        logger.info("Archiving disabled by list policy for %s", list_name)
        return
    if "Message-Id" not in message:
        raise ValueError("No 'Message-Id' header in email", message)
    # timeit("2 after ml, before checking email & sender")
    msg_id = get_message_id(message)
    if Email.objects.filter(mailinglist=mlist, message_id=msg_id).exists():
        raise DuplicateMessage(msg_id)
    email = Email(mailinglist=mlist, message_id=msg_id)
    email.in_reply_to = get_ref(message)  # Find thread id
    if message.get_unixfrom() is not None:
        mo = UNIXFROM_DATE_RE.match(message.get_unixfrom())
        if mo:
            archived_date = parsedate(mo.group(1))
            if archived_date is not None:
                email.archived_date = archived_date

    # Sender
    try:
        from_str = header_to_unicode(message['From'])
        from_name, from_email = parseaddr(from_str)
        from_name = from_name.strip()
        sender_address = from_email.encode('ascii').decode("ascii").strip()
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise ValueError("Non-ascii sender address", message)
    if not sender_address:
        if from_name:
            sender_address = re.sub("[^a-z0-9]", "", from_name.lower())
            if not sender_address:
                sender_address = "unknown"
            sender_address = "{}@example.com".format(sender_address)
        else:
            sender_address = "unknown@example.com"
    email.sender_name = from_name
    sender = Sender.objects.get_or_create(address=sender_address)[0]
    email.sender = sender
    if not getattr(settings, "HYPERKITTY_BATCH_MODE", False):
        sender_mailman_id(sender.pk)
    # timeit("3 after sender, before email content")

    # Headers
    email.subject = header_to_unicode(message.get('Subject'))
    if email.subject is not None:
        # limit subject size to 512, it's a varchar field
        email.subject = email.subject[:512]
    msg_date = parsedate(message.get("Date"))
    if msg_date is None:
        # Absent or unparseable date
        msg_date = timezone.now()
    utcoffset = msg_date.utcoffset()
    if msg_date.tzinfo is not None:
        msg_date = msg_date.astimezone(timezone.utc)  # store in UTC
    email.date = msg_date
    if utcoffset is None:
        email.timezone = 0
    else:
        # in minutes
        email.timezone = int(
            ((utcoffset.days * 24 * 60 * 60) + utcoffset.seconds) / 60)

    # Content
    scrubber = Scrubber(message)
    # warning: scrubbing modifies the msg in-place
    email.content, attachments = scrubber.scrub()
    # timeit("4 after email content, before signals")

    # TODO: detect category?

    # Find the parent email.
    # This can't be moved to Email.on_pre_save() because Email.set_parent()
    # needs to be free to change the parent independently from the in_reply_to
    # property, and will save() the instance.
    # This, along with some of the work done in Email.on_pre_save(), could be
    # moved to an async task, but the rest of the app must be able to cope with
    # emails lacking this data, and email being process randomly (child before
    # parent). The work in Email.on_post_created() also depends on it, so be
    # careful with task dependencies if you ever do this.
    # Plus, it has "premature optimization" written all over it.
    if email.in_reply_to is not None:
        try:
            ref_msg = Email.objects.get(
                mailinglist=email.mailinglist,
                message_id=email.in_reply_to)
        except Email.DoesNotExist:
            # the parent may not be archived (on partial imports), create a new
            # thread for now.
            pass
        else:
            # re-use parent's thread-id
            email.parent = ref_msg
            email.thread_id = ref_msg.thread_id

    try:
        email.save()
    except DataError as e:
        raise ValueError(str(e))

    # Attachments (email must have been saved before)
    for attachment in attachments:
        counter, name, content_type, encoding, content = attachment
        if Attachment.objects.filter(email=email, counter=counter).exists():
            continue
        att = Attachment.objects.create(
            email=email, counter=counter, name=name, content_type=content_type,
            encoding=encoding)
        att.set_content(content)
        att.save()

    return email.message_id_hash



# Generated by Django 2.0.8 on 2019-01-26 16:25

from django.db import migrations

import hyperkitty.models.category


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0017_file_attachments'),
    ]

    operations = [
        migrations.AlterField(
            model_name='threadcategory',
            name='color',
            field=hyperkitty.models.category.ColorField(max_length=7),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import re
from smtplib import SMTPResponseException
from urllib.error import HTTPError

from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.core.mail import EmailMessage

from django_mailman3.lib.mailman import get_subscriptions
from mailmanclient import MailmanConnectionError

from hyperkitty.lib import mailman


class PostingFailed(Exception):
    pass


def get_sender(request, mlist):
    """Returns the appropriate sender email address"""
    if not request.user.is_authenticated:
        return None
    # Fallback to the logged-in user
    address = request.user.email
    # Try to get the email used to susbscribe to the list
    subscriptions = get_subscriptions(request.user)
    if mlist.list_id in subscriptions:
        address = subscriptions[mlist.list_id]
    return str(address)


def get_from(request, address):
    """Returns the appropriate 'From' header"""
    assert address is not None
    display_name = "%s %s" % (request.user.first_name, request.user.last_name)
    # Get the display_name from the Address in Mailman? And if not found,
    # from the User in Mailman?
    if display_name.strip():
        return '"%s" <%s>' % (display_name, address)
    else:
        return address


def post_to_list(request, mlist, subject, message, headers=None,
                 attachments=None):
    if not mlist:
        # Make sure the list exists to avoid posting to any email addess
        raise SuspiciousOperation("I don't know this mailing-list")
    if headers is None:
        headers = {}

    sender = headers.pop("From", get_sender(request, mlist))
    display_name = "%s %s" % (request.user.first_name, request.user.last_name)
    if display_name.strip():
        from_email = '"%s" <%s>' % (display_name, sender)
    else:
        from_email = sender
    # Unwrap and collapse spaces
    subject = re.sub(r'\n+', ' ', subject)
    subject = re.sub(r'\s+', ' ', subject)

    # Check that the user is subscribed
    try:
        subscribed_now = mailman.subscribe(
            mlist.list_id, request.user, sender, display_name)
    except MailmanConnectionError:
        raise PostingFailed("Can't connect to Mailman's REST server, "
                            "your message has not been sent.")
    except HTTPError as e:
        raise PostingFailed("Failed to post message: %s", str(e))

    # send the message
    headers["User-Agent"] = (
        "HyperKitty on %s" % request.build_absolute_uri("/"))
    msg = EmailMessage(
               subject=subject,
               body=message,
               from_email=from_email,
               to=[mlist.name],
               headers=headers,
               )
    # Attachments
    if attachments:
        if not isinstance(attachments, list):
            attachments = [attachments]
        for attach in attachments:
            msg.attach(attach.name, attach.read())
    # XXX: Inject into the incoming queue instead?
    if not settings.DEBUG:
        # Don't send mail in debug mode, just in case...
        try:
            msg.send()
        except SMTPResponseException as e:
            raise PostingFailed('Message not sent: SMTP error {} {}'.format(
                                e.smtp_code, e.smtp_error))
    return subscribed_now


def reply_subject(subject):
    if not subject.lower().startswith("re:"):
        return "Re: %s" % subject
    else:
        return subject



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
import gzip
import mailbox
import os
import tempfile
from io import StringIO

from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse

from hyperkitty.lib.compat import get_list_by_name, month_name_to_num
from hyperkitty.models import Email, MailingList


def summary(request, list_name=None):
    if list_name is None:
        return redirect(reverse('hk_root'))
    mlist = get_list_by_name(list_name, request.get_host())
    return redirect(reverse('hk_list_overview',
                            kwargs={'mlist_fqdn': mlist.name}))


def arch_month(request, list_name, year, month_name, summary_type="thread"):
    mlist = get_list_by_name(list_name, request.get_host())
    return redirect(reverse('hk_archives_with_month', kwargs={
        'mlist_fqdn': mlist.name,
        'year': year,
        'month': str(month_name_to_num(month_name)).rjust(2, "0"),
        }))


def arch_month_mbox(request, list_name, year, month_name):
    """
    The messages must be rebuilt before being added to the mbox file, including
    headers and the textual content, making sure to escape email addresses.
    """
    return HttpResponse("Not implemented yet.",
                        content_type="text/plain", status=500)
    mlist = get_list_by_name(list_name, request.get_host())
    month = month_name_to_num(month_name)
    year = int(year)
    begin_date = datetime.datetime(year, month, 1)
    if month != 12:
        end_month = month + 1
    else:
        end_month = 1
    end_date = datetime.datetime(year, end_month, 1)
    messages = Email.objects.filter(
        mailinglist=mlist, date__gte=begin_date, date__lte=end_date
        ).order_by("date")
    mboxfile, mboxfilepath = tempfile.mkstemp(
        prefix="hyperkitty-", suffix=".mbox.gz")
    os.close(mboxfile)
    mbox = mailbox.mbox(mboxfilepath)
    for msg in messages:
        mbox.add(msg.full)
    mbox.close()
    content = StringIO()
    zipped_content = gzip.GzipFile(fileobj=content)
    with gzip.GzipFile(fileobj=content, mode="wb") as zipped_content:
        with open(mboxfilepath, "rb") as mboxfile:
            zipped_content.write(mboxfile.read())
    response = HttpResponse(content.getvalue())
    content.close()
    response['Content-Type'] = "application/mbox+gz"
    response['Content-Disposition'] = 'attachment; filename=%d-%s.txt.gz' \
        % (year, month_name)
    response['Content-Length'] = len(response.content)
    os.remove(mboxfilepath)
    return response


def message(request, list_name, year, month_name, msg_num):
    mlist = get_list_by_name(list_name, request.get_host())
    msg_num = int(msg_num) - 1  # pipermail starts at 1, not 0
    if msg_num < 0:
        raise Http404("No such message in this mailing-list.")
    try:
        msg = Email.objects.filter(
            mailinglist=mlist).order_by("archived_date")[msg_num]
    except IndexError:
        raise Http404("No such message in this mailing-list.")
    return redirect(reverse('hk_message_index', kwargs={
        'mlist_fqdn': mlist.name,
        'message_id_hash': msg.message_id_hash,
        }))


def redirect_list_id(request, list_id):
    mlist = get_object_or_404(MailingList, list_id=list_id)
    url = request.path.replace(list_id, mlist.name, 1)
    if request.GET:
        url = "%s?%s" % (url, request.GET.urlencode())
    return redirect(url)


def redirect_lists(request):
    url = request.path.replace("/lists/", "/list/", 1)
    if request.GET:
        url = "%s?%s" % (url, request.GET.urlencode())
    return redirect(url)



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#


from django.urls import reverse

from rest_framework import serializers

from hyperkitty.models import Attachment
from .utils import MLChildHyperlinkedRelatedField


class AttachmentSerializer(serializers.HyperlinkedModelSerializer):
    email = MLChildHyperlinkedRelatedField(
        view_name='hk_api_email_detail', read_only=True,
        lookup_field="message_id_hash")
    download = serializers.SerializerMethodField()

    class Meta:
        model = Attachment
        fields = (
            "email", "counter", "name", "content_type", "encoding",
            "size", "download",
        )

    def get_download(self, obj):
        relative_url = reverse(
            "hk_message_attachment", kwargs=dict(
                mlist_fqdn=obj.email.mailinglist.name,
                message_id_hash=obj.email.message_id_hash,
                counter=obj.counter,
                filename=obj.name,
            )
        )
        if "request" in self.context:
            return self.context["request"].build_absolute_uri(relative_url)
        else:
            return relative_url



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
import json
import re
from collections import OrderedDict

from django import template
from django.conf import settings
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.timezone import utc

from dateutil.tz import tzoffset
from django_gravatar.templatetags.gravatar import gravatar as gravatar_orig

import hyperkitty.lib.posting
from hyperkitty.lib.utils import stripped_subject


register = template.Library()

MAILTO_RE = re.compile(
    "<a href=['\"]mailto:([^'\"]+)@([^'\"]+)['\"]>[^<]+</a>")
SNIPPED_RE = re.compile(r"^(\s*&gt;).*$", re.M)
SNIPPED_BEGIN_PGP = re.compile("^.*(BEGIN PGP SIGNATURE).*$", re.M)
SNIPPED_END_PGP = re.compile("^.*(END PGP SIGNATURE).*$", re.M)


@register.filter(name='sort')
def listsort(value):
    if isinstance(value, dict):
        new_dict = OrderedDict()
        key_list = sorted(value.keys())
        key_list.reverse()
        for key in key_list:
            values = sorted(value[key])
            values.reverse()
            new_dict[key] = values
        return new_dict.items()
    elif isinstance(value, list):
        return sorted(value)
    else:
        return value
    listsort.is_safe = True


@register.filter(name="monthtodate")
def to_date(month, year):
    return datetime.date(year, month, 1)


# From http://djangosnippets.org/snippets/1259/
@register.filter
def truncatesmart(value, limit=80):
    """
    Truncates a string after a given number of chars keeping whole words.

    Usage:
        {{ string|truncatesmart }}
        {{ string|truncatesmart:50 }}
    """
    try:
        limit = int(limit)
    # invalid literal for int()
    except ValueError:
        # Fail silently.
        return value

    # Return the string itself if length is smaller or equal to the limit
    if len(value) <= limit:
        return value

    # Cut the string
    value = value[:limit]

    # Break into words and remove the last
    words = value.split(' ')[:-1]

    # Join the words and return
    return ' '.join(words) + '...'


@register.filter(is_safe=True)
def escapeemail(text):
    """To escape email addresses"""
    return text.replace("@", "\uff20")


@register.filter(is_safe=True)
def escapeemaillinks(text):
    """To escape email addresses in links"""
    # reverse the effect of urlize() on email addresses
    return MAILTO_RE.sub(r"\1(a)\2", text)


@register.filter()
def date_with_senders_timezone(email):
    """
    Rebuild the date of an email taking the sender's timezone into account.
    """
    # tzoffset requires seconds
    tz = tzoffset(None, email.timezone * 60)
    return email.date.astimezone(tz)


@register.filter(needs_autoescape=True)
def snip_quoted(content, quotemsg="...",
                autoescape=None, add_switch=True, quote_index=0):
    """Snip quoted text in messages.

    This method recursively tries to find quoted text in an email and return
    HTML with proper quotes. We process only 4 levels of indent so as to
    prevent too much processing of text.

    :param content: The content to be quoted.
    :param quotemsg: The content of the switch which shows and hides the
        quoted text.
    :param add_switch: Boolean specifying if we should add the switch to
         show-hide quoted content. To prevent having too many switches,
         we only add a single switch to the top-level quoted text.
    :param quote_index: Which level of nested quote we are in. This is
         incremented on a recursive call and we limit it to a maximum
         of 3, starting from 0.

    """
    # We don't want to go into infinite loop of quoting, so we limit the
    # quoting to 0-3, 4 levels only.

    # XXX(maxking): This function is currently unused since we switched to
    # using markdown rendering from decorate.py. I am going to keep this around
    # for a white till we are sure using the markdown renderer is good to
    # replace this.
    if quote_index > 3 or quote_index < 0:
        return content
    if autoescape:
        content = conditional_escape(content)
    quoted = []
    current_quote = []
    current_quote_orig = []
    for line in content.split("\n"):
        match = SNIPPED_RE.match(line)
        if match is not None:
            current_quote_orig.append(line)
            content_start = len(match.group(1))
            current_quote.append(line[content_start:])
        else:
            if current_quote_orig:
                current_quote_orig.append("")
                quoted.append((current_quote_orig[:], current_quote[:]))
                current_quote = []
                current_quote_orig = []

    for quote_orig, quote in quoted:
        replaced = ''
        # Add a switch if we are needed to, usually, this is used for the first
        # level of quoting. Any further nested levels do not require a switch.
        if add_switch:
            replaced = (
                '<div class="quoted-switch">'
                '<a style="font-weight:normal" href="#">{quotemsg}</a>'
                '</div>').format(quotemsg=quotemsg)
        # Finally, generate the quoted content. We recursively call
        # `snip_quoted` to process further levels of quotes in the message.
        replaced += (
            '<div class="quoted-text quoted-text-{index}"> {quote} </div>'
        ).format(
            quote=snip_quoted("\n".join(quote),
                              add_switch=False, quote_index=quote_index+1),
            index=quote_index)
        # Replace the original content with the new one.
        content = content.replace("\n".join(quote_orig), replaced)
    return mark_safe(content)


@register.filter(needs_autoescape=True)
def snip_pgp(content, quotemsg="...PGP SIGNATURE...", autoescape=None):
    """Snip pgp signature in messages"""
    if autoescape:
        content = conditional_escape(content)
    quoted = []
    current_quote = []
    current_quote_orig = []
    pgp_signature = False
    for line in content.split("\n"):
        match_start = SNIPPED_BEGIN_PGP.match(line)
        match_end = SNIPPED_END_PGP.match(line)
        if match_end is not None or match_start is not None or pgp_signature:
            current_quote_orig.append(line)
            current_quote.append(line)
        # start of pgp signature
        if match_start is not None:
            pgp_signature = True
        # end of pgp signature
        elif match_end is not None:
            pgp_signature = False
            if current_quote_orig:
                current_quote_orig.append("")
                quoted.append((current_quote_orig[:], current_quote[:]))
                current_quote = []
                current_quote_orig = []
    for quote_orig, quote in quoted:
        replaced = (
            '<div class="quoted-switch">'
            '<a href="#" class="pgp">{quotemsg}</a></div>'
            '<div class="quoted-text">{quote} </div>'
            ).format(quotemsg=quotemsg, quote="\n".join(quote))
        content = content.replace("\n".join(quote_orig), replaced)
    return mark_safe(content)


@register.filter()
def multiply(num1, num2):
    if int(num2) == float(num2):
        num2 = int(num2)
    else:
        num2 = float(num2)
    return num1 * num2


@register.simple_tag(takes_context=True)
def is_message_new(context, refdate):
    user = context["user"]
    if "last_view" not in context:
        return False  # viewing a single message
    last_view = context["last_view"]
    refdate = refdate.replace(tzinfo=utc)
    return (
        user.is_authenticated and
        (not last_view or refdate > last_view)
        )


@register.filter
def until(value, limit):
    return value.partition(limit)[0]


@register.filter
def to_json(value):
    return json.dumps(value)


@register.filter
def get_item(dictionary, key):
    return dictionary.get(key)


@register.filter(is_safe=True)
def num_comments(thread):
    """Returns the number of comments in a thread"""
    try:
        return thread.emails_count - 1
    except (ValueError, TypeError):
        return ''


@register.filter
def reply_subject(value):
    return hyperkitty.lib.posting.reply_subject(value)


@register.filter(name="strip_subject")
def strip_subject(subject, mlist):
    return stripped_subject(mlist, subject)


@register.filter
def is_unread_by(thread, user):
    return thread.is_unread_by(user)


@register.filter
def sort_by_name(p_list):
    return sorted(p_list, key=lambda p: p.name and p.name.lower())


@register.simple_tag()
def gravatar(*args, **kw):
    """A proxy for django-gravatar's template.

    This templatetag allows disabling Gravatar altogether using
    HYPERKITTY_ENABLE_GRAVATAR setting, which is True by default.
    """
    if not getattr(settings, 'HYPERKITTY_ENABLE_GRAVATAR', True):
        return mark_safe('')
    return gravatar_orig(*args, **kw)


@register.simple_tag()
def settings_value_equals(name, value):
    """Get the settings value to use in templates.

    Default value is set to empty string, which might not work for all settings
    types.
    """
    return getattr(settings, name, '') == value


@register.simple_tag()
def export_allowed():
    """Returns the HYPERKITTY_MBOX_EXPORT settings value. Defaults to True."""
    return getattr(settings, 'HYPERKITTY_MBOX_EXPORT', True)



# -*- coding: utf-8 -*-
# flake8: noqa

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0011_email_parent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='thread_order',
            field=models.IntegerField(db_index=True, null=True, blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>

"""
Move attachments from database to file system after setting
HYPERKITTY_ATTACHMENT_FOLDER.
"""

import os

from django.core.management.base import BaseCommand, CommandError

from hyperkitty.management.utils import setup_logging
from hyperkitty.models.email import Attachment


class Command(BaseCommand):
    help = """Move attachments from database to file system after setting
HYPERKITTY_ATTACHMENT_FOLDER."""

    def add_arguments(self, parser):
        for action in parser._actions:
            for option in ('-v', '--verbosity'):
                if vars(action)['option_strings'][0] == option:
                    parser._handle_conflict_resolve(
                        None, [(option, action)])
        parser.add_argument(
            '-v', '--verbosity', default=0,
            type=int, choices=[0, 1],
            help="""Verbosity = 1 will print a dot for each 100 attachments
moved."""
        )
        parser.add_argument(
            '-c', '--chunk-size', default=100, type=int,
            help="""Specify the number of attachments to retrieve at one time
from the database. Default is 100. Larger values use more memory."""
        )

    def handle(self, *args, **options):
        options["verbosity"] = int(options.get("verbosity", "0"))
        options["chunk-size"] = int(options.get("chunk-size", 100))
        setup_logging(self, options["verbosity"])
        if args:
            raise CommandError("no arguments allowed")
        count = 0
        for attachment in Attachment.objects.iterator(
                chunk_size=options["chunk-size"]):
            path = attachment._get_folder()
            if path is None:
                raise CommandError('HYPERKITTY_ATTACHMENT_FOLDER is not set')
            if attachment.content is None:
                continue
            count += 1
            if options['verbosity'] > 0:
                if count % 100 == 0:
                    print('.', end='', flush=True)
                if count % 7000 == 0:
                    print()
            if not os.path.exists(path):
                os.makedirs(path)
            file = os.path.join(path, str(attachment.counter))
            with open(file, 'wb') as fp:
                fp.write(bytes(attachment.content))
            attachment.content = None
            attachment.save()
        if options['verbosity'] > 0:
            print()
        print(f'{count} attachments moved.')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import datetime
from functools import wraps

from django.http import Http404
from django.shortcuts import render
from django.utils.timezone import utc

from django_mailman3.lib.mailman import get_subscriptions

from hyperkitty.forms import CategoryForm
from hyperkitty.lib.posting import get_sender
from hyperkitty.models import MailingList, ThreadCategory


def get_months(mlist):
    """Return a dictionnary of years, months for which there are
    potentially archives available for a given list (based on the
    oldest post on the list).

    :param mlist: name of the mailing list in which this email should be
    searched.
    """
    date_first = mlist.cached_values["first_date"]()
    now = datetime.datetime.now()
    if not date_first:
        # No messages on this list, return the current month.
        return {now.year: [now.month]}
    archives = {}
    year = date_first.year
    month = date_first.month
    while year < now.year:
        archives[year] = range(1, 13)[(month-1):]
        year = year + 1
        month = 1
    archives[now.year] = range(1, 13)[:now.month]
    return archives


def get_display_dates(year, month, day):
    if day is None:
        start_day = 1
    else:
        start_day = int(day)
    begin_date = datetime.datetime(
        int(year), int(month), start_day, tzinfo=utc)

    if day is None:
        end_date = begin_date + datetime.timedelta(days=32)
        end_date = end_date.replace(day=1)
    else:
        end_date = begin_date + datetime.timedelta(days=1)

    return begin_date, end_date


def daterange(start_date, end_date):
    for n in range(int((end_date - start_date).days)):
        yield start_date + datetime.timedelta(n)


def get_category_widget(request, current_category=None):
    """
    Returns the category form and the applicable category object (or None if no
    category is set for this thread).

    If current_category is not provided or None, try to deduce it from the POST
    request.
    """

    categories = [
        (c.name, c.name.upper())
        for c in ThreadCategory.objects.all()
        ] + [("", "no category")]
    if request.method == "POST":
        category_form = CategoryForm(request.POST)
    else:
        category_form = CategoryForm(
            initial={"category": current_category or ""})
    category_form["category"].field.choices = categories

    if request.method == "POST" and category_form.is_valid():
        # is_valid() must be called after the choices have been set
        current_category = category_form.cleaned_data["category"]

    if not current_category:
        category = None
    else:
        try:
            category = ThreadCategory.objects.get(name=current_category)
        except ThreadCategory.DoesNotExist:
            category = None
    return category, category_form


# View decorator: check that the list is authorized
def check_mlist_private(func):
    @wraps(func)
    def inner(request, *args, **kwargs):
        if "mlist_fqdn" in kwargs:
            mlist_fqdn = kwargs["mlist_fqdn"]
        else:
            mlist_fqdn = args[0]
        try:
            mlist = MailingList.objects.get(name=mlist_fqdn)
        except MailingList.DoesNotExist:
            raise Http404("No archived mailing-list by that name.")
        if not is_mlist_authorized(request, mlist):
            return render(
                request, "hyperkitty/errors/private.html", {
                    "mlist": mlist,
                }, status=403)
        # Set this here so that the view doesn't have to fetch and create
        # the mlist object again.
        request.mlist = mlist
        return func(request, *args, **kwargs)
    return inner


def is_mlist_authorized(request, mlist):
    if not mlist.is_private:
        return True
    if request.user.is_superuser:
        return True
    if not request.user.is_authenticated:
        return False
    # Private list and logged-in user: check subscriptions
    if mlist.list_id in get_subscriptions(request.user):
        return True
    else:
        return False


def get_posting_form(formclass, request, mlist, data=None):
    form = formclass(data, initial={
        "sender": get_sender(request, mlist)})
    if request.user.is_authenticated:
        form.fields['sender'].choices = [
            (a, a) for a in request.user.hyperkitty_profile.addresses]
    return form



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404

from rest_framework import generics, serializers
from rest_framework.pagination import LimitOffsetPagination

from hyperkitty.lib.view_helpers import is_mlist_authorized
from hyperkitty.models import MailingList, Thread
from .utils import (
    IsMailingListPublicOrIsMember, MLChildHyperlinkedRelatedField)


class ThreadShortSerializer(serializers.HyperlinkedModelSerializer):
    url = MLChildHyperlinkedRelatedField(
        view_name='hk_api_thread_detail', read_only=True,
        lookup_field="thread_id", source="*")
    mailinglist = serializers.HyperlinkedRelatedField(
        view_name='hk_api_mailinglist_detail', read_only=True,
        lookup_field="name", lookup_url_kwarg="mlist_fqdn")
    starting_email = MLChildHyperlinkedRelatedField(
        view_name='hk_api_email_detail', read_only=True,
        lookup_field="message_id_hash")
    votes_total = serializers.IntegerField(min_value=0)
    emails = MLChildHyperlinkedRelatedField(
        view_name='hk_api_thread_email_list', read_only=True,
        lookup_field="thread_id", source="*")
    replies_count = serializers.SerializerMethodField()
    next_thread = MLChildHyperlinkedRelatedField(
        view_name='hk_api_thread_detail', read_only=True,
        lookup_field="thread_id")
    prev_thread = MLChildHyperlinkedRelatedField(
        view_name='hk_api_thread_detail', read_only=True,
        lookup_field="thread_id")

    class Meta:
        model = Thread
        fields = ("url", "mailinglist", "thread_id", "subject", "date_active",
                  "starting_email", "emails", "votes_total",
                  "replies_count", "next_thread", "prev_thread")

    def get_replies_count(self, obj):
        return obj.emails_count - 1


class ThreadSerializer(ThreadShortSerializer):
    votes = serializers.SerializerMethodField()
    participants = serializers.SerializerMethodField()

    class Meta:
        model = Thread
        fields = ThreadShortSerializer.Meta.fields + (
            "votes", "participants", "participants_count",
            )

    def get_votes(self, obj):
        return obj.get_votes()

    def get_participants(self, obj):
        return [
            {"name": p[0].replace("@", " (a) "),
             "email": p[1].replace("@", " (a) "),
             }
            for p in obj.participants
            ]


class ThreadList(generics.ListAPIView):
    """List threads"""

    serializer_class = ThreadShortSerializer
    pagination_class = LimitOffsetPagination
    ordering = ("-date_active", )

    def get_queryset(self):
        mlist = MailingList.objects.get(name=self.kwargs["mlist_fqdn"])
        if not is_mlist_authorized(self.request, mlist):
            raise PermissionDenied
        return Thread.objects.filter(
                mailinglist__name=self.kwargs["mlist_fqdn"],
            ).order_by("-date_active")


class ThreadDetail(generics.RetrieveAPIView):
    """Show a thread"""

    serializer_class = ThreadSerializer
    pagination_class = LimitOffsetPagination
    permission_classes = [IsMailingListPublicOrIsMember]

    def get_object(self):
        thread = get_object_or_404(
            Thread,
            mailinglist__name=self.kwargs["mlist_fqdn"],
            thread_id=self.kwargs["thread_id"],
            )
        return thread



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404

from rest_framework import generics, serializers
from rest_framework.pagination import LimitOffsetPagination

from hyperkitty.lib.view_helpers import is_mlist_authorized
from hyperkitty.models import ArchivePolicy, Email, MailingList
from .attachment import AttachmentSerializer
from .sender import SenderSerializer
from .utils import (
    IsMailingListPublicOrIsMember, MLChildHyperlinkedRelatedField)


class EmailShortSerializer(serializers.HyperlinkedModelSerializer):
    url = MLChildHyperlinkedRelatedField(
        view_name='hk_api_email_detail', read_only=True,
        lookup_field="message_id_hash", source="*")
    mailinglist = serializers.HyperlinkedRelatedField(
        view_name='hk_api_mailinglist_detail', read_only=True,
        lookup_field="name", lookup_url_kwarg="mlist_fqdn")
    thread = MLChildHyperlinkedRelatedField(
        view_name='hk_api_thread_detail', read_only=True,
        lookup_field="thread_id")
    parent = MLChildHyperlinkedRelatedField(
        view_name='hk_api_email_detail', read_only=True,
        lookup_field="message_id_hash")
    children = MLChildHyperlinkedRelatedField(
        view_name='hk_api_email_detail', read_only=True,
        lookup_field="message_id_hash", many=True)
    sender = SenderSerializer()

    class Meta:
        model = Email
        fields = ("url", "mailinglist", "message_id", "message_id_hash",
                  "thread", "sender", "sender_name", "subject", "date",
                  "parent", "children",
                  )


class EmailSerializer(EmailShortSerializer):
    votes = serializers.SerializerMethodField()
    attachments = AttachmentSerializer(many=True, read_only=True)

    class Meta:
        model = Email
        fields = EmailShortSerializer.Meta.fields + (
            "votes", "content", "attachments")

    def get_votes(self, obj):
        return obj.get_votes()


class EmailList(generics.ListAPIView):
    """List emails"""

    serializer_class = EmailShortSerializer
    ordering_fields = ("archived_date", "thread_order", "date")
    pagination_class = LimitOffsetPagination

    def get_queryset(self):
        mlist = MailingList.objects.get(name=self.kwargs["mlist_fqdn"])
        if not is_mlist_authorized(self.request, mlist):
            raise PermissionDenied
        query = Email.objects.filter(
                mailinglist__name=self.kwargs["mlist_fqdn"])
        if "thread_id" in self.kwargs:
            query = query.filter(
                    thread__thread_id=self.kwargs["thread_id"]
                ).order_by("thread_order")
        else:
            query = query.order_by("-archived_date")
        return query


class EmailListBySender(generics.ListAPIView):
    """List emails by sender"""

    serializer_class = EmailShortSerializer
    pagination_class = LimitOffsetPagination

    def get_queryset(self):
        key = self.kwargs["mailman_id"]
        query = Email.objects.exclude(
            mailinglist__archive_policy=ArchivePolicy.private.value
        )
        if "@" in key:
            query = query.filter(sender__address=key)
        else:
            query = query.filter(sender__mailman_id=key)
        return query.order_by("-archived_date")


class EmailDetail(generics.RetrieveAPIView):
    """Show an email"""

    serializer_class = EmailSerializer
    permission_classes = [IsMailingListPublicOrIsMember]

    def get_object(self):
        email = get_object_or_404(
            Email,
            mailinglist__name=self.kwargs["mlist_fqdn"],
            message_id_hash=self.kwargs["message_id_hash"],
            )
        return email



# Generated by Django 3.0.6 on 2020-06-19 05:23

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('hyperkitty', '0020_auto_20190907_1927'),
    ]

    operations = [
        migrations.AddField(
            model_name='mailinglist',
            name='moderators',
            field=models.ManyToManyField(
                related_name='moderated_lists', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='mailinglist',
            name='owners',
            field=models.ManyToManyField(
                related_name='owned_lists', to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

import logging
from urllib.error import HTTPError

from django.db import models

from django_mailman3.lib.mailman import get_mailman_client
from mailmanclient import MailmanConnectionError


logger = logging.getLogger(__name__)


class Sender(models.Model):
    address = models.EmailField(max_length=255, primary_key=True)
    mailman_id = models.CharField(max_length=255, null=True, db_index=True)

    def __str__(self):
        return self.address

    @property
    def name(self):
        try:
            return self.emails.order_by("-date").values_list(
                "sender_name", flat=True)[0]
        except IndexError:
            return "(no name)"

    @property
    def names(self):
        return self.emails.order_by("-date").values_list(
            "sender_name", flat=True)

    def set_mailman_id(self):
        try:
            client = get_mailman_client()
            mm_user = client.get_user(self.address)
        except HTTPError as e:
            if e.code == 404:
                return  # User not found in Mailman
            # normalize all possible error types
            raise MailmanConnectionError(e)
        except ValueError:
            # This smells like a badly formatted email address (saw it in the
            # wild)
            logger.warning(
                "Invalid response when getting user %s from Mailman",
                self.address)
            return  # Ignore it
        self.mailman_id = mm_user.user_id
        self.save()
        # # Go further and associate the user's other addresses?
        # Sender.objects.filter(address__in=mm_user.addresses
        #     ).update(mailman_id=mm_user.user_id)



# -*- coding: utf-8 -*-
# flake8: noqa
import zoneinfo

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


TIMEZONES = sorted([(tz, tz) for tz in zoneinfo.available_timezones()])


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('counter', models.SmallIntegerField()),
                ('name', models.CharField(max_length=255)),
                ('content_type', models.CharField(max_length=255)),
                ('encoding', models.CharField(max_length=255, null=True)),
                ('size', models.IntegerField()),
                ('content', models.BinaryField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('message_id', models.CharField(max_length=255, db_index=True)),
                ('message_id_hash', models.CharField(max_length=255, db_index=True)),
                ('subject', models.CharField(max_length='512', db_index=True)),
                ('content', models.TextField()),
                ('date', models.DateTimeField(db_index=True)),
                ('timezone', models.SmallIntegerField()),
                ('in_reply_to', models.CharField(max_length=255, null=True, blank=True)),
                ('archived_date', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('thread_depth', models.IntegerField(default=0)),
                ('thread_order', models.IntegerField(default=0, db_index=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Favorite',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='LastView',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('view_date', models.DateTimeField(auto_now=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='MailingList',
            fields=[
                ('name', models.CharField(max_length=254, serialize=False, primary_key=True)),
                ('display_name', models.CharField(max_length=255)),
                ('description', models.TextField()),
                ('subject_prefix', models.CharField(max_length=255)),
                ('archive_policy', models.IntegerField(default=2, choices=[(0, 'never'), (1, 'private'), (2, 'public')])),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('karma', models.IntegerField(default=1)),
                ('timezone', models.CharField(default='', max_length=100, choices=TIMEZONES)),
                ('user', models.OneToOneField(related_name='hyperkitty_profile', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Sender',
            fields=[
                ('address', models.EmailField(max_length=255, serialize=False, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('mailman_id', models.CharField(max_length=255, null=True, db_index=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=255, db_index=True)),
            ],
            options={
                'ordering': ['name'],
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Tagging',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('tag', models.ForeignKey(to='hyperkitty.Tag', on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Thread',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('thread_id', models.CharField(max_length=255, db_index=True)),
                ('date_active', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ThreadCategory',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=255, db_index=True)),
                ('color', models.CharField(max_length=7)),
            ],
            options={
                'verbose_name_plural': 'Thread categories',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Vote',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('value', models.SmallIntegerField(db_index=True)),
                ('email', models.ForeignKey(related_name='votes', to='hyperkitty.Email', on_delete=models.CASCADE)),
                ('user', models.ForeignKey(related_name='votes', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='vote',
            unique_together=set([('email', 'user')]),
        ),
        migrations.AddField(
            model_name='thread',
            name='category',
            field=models.ForeignKey(related_name='threads', to='hyperkitty.ThreadCategory', null=True, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='thread',
            name='mailinglist',
            field=models.ForeignKey(related_name='threads', to='hyperkitty.MailingList', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='thread',
            unique_together=set([('mailinglist', 'thread_id')]),
        ),
        migrations.AddField(
            model_name='tagging',
            name='thread',
            field=models.ForeignKey(to='hyperkitty.Thread', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='tagging',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='tag',
            name='threads',
            field=models.ManyToManyField(related_name='tags', through='hyperkitty.Tagging', to='hyperkitty.Thread'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='tag',
            name='users',
            field=models.ManyToManyField(related_name='tags', through='hyperkitty.Tagging', to=settings.AUTH_USER_MODEL),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='lastview',
            name='thread',
            field=models.ForeignKey(related_name='lastviews', to='hyperkitty.Thread', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='lastview',
            name='user',
            field=models.ForeignKey(related_name='lastviews', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='favorite',
            name='thread',
            field=models.ForeignKey(related_name='favorites', to='hyperkitty.Thread', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='favorite',
            name='user',
            field=models.ForeignKey(related_name='favorites', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='email',
            name='mailinglist',
            field=models.ForeignKey(related_name='emails', to='hyperkitty.MailingList', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='email',
            name='parent',
            field=models.ForeignKey(related_name='children',
                                    on_delete=django.db.models.deletion.SET_NULL,
                                    blank=True, to='hyperkitty.Email', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='email',
            name='sender',
            field=models.ForeignKey(related_name='emails', to='hyperkitty.Sender', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='email',
            name='thread',
            field=models.ForeignKey(related_name='emails', to='hyperkitty.Thread', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='email',
            unique_together=set([('mailinglist', 'message_id')]),
        ),
        migrations.AddField(
            model_name='attachment',
            name='email',
            field=models.ForeignKey(related_name='attachments', to='hyperkitty.Email', on_delete=models.CASCADE),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='attachment',
            unique_together=set([('email', 'counter')]),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>

"""
Sync properties from Mailman into HyperKitty
"""

from django.core.management.base import BaseCommand, CommandError

from hyperkitty.lib.mailman import sync_with_mailman
from hyperkitty.management.utils import setup_logging


class Command(BaseCommand):
    help = "Sync properties from Mailman into HyperKitty"

    def add_arguments(self, parser):
        parser.add_argument(
            '--overwrite', action='store_true', default=False,
            help="overwrite existing Mailman IDs in HyperKitty's database")

    def handle(self, *args, **options):
        options["verbosity"] = int(options.get("verbosity", "1"))
        setup_logging(self, options["verbosity"])
        if args:
            raise CommandError("no arguments allowed")
        sync_with_mailman(overwrite=options.get("overwrite", False))



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from rest_framework import permissions, serializers

from hyperkitty.lib.view_helpers import is_mlist_authorized
from hyperkitty.models import MailingList


class MLChildHyperlinkedRelatedField(serializers.HyperlinkedRelatedField):

    def get_url(self, obj, view_name, request, format):
        lookup_value = getattr(obj, self.lookup_field)
        kwargs = {self.lookup_url_kwarg: lookup_value,
                  "mlist_fqdn": obj.mailinglist.name}
        return self.reverse(view_name, kwargs=kwargs, request=request,
                            format=format)

    def get_object(self, view_name, view_args, view_kwargs):
        lookup_value = view_kwargs[self.lookup_url_kwarg]
        lookup_kwargs = {self.lookup_field: lookup_value,
                         "mailinglist__name": view_kwargs["mlist_fqdn"]}
        return self.get_queryset().get(**lookup_kwargs)


class EnumField(serializers.IntegerField):

    def __init__(self, **kwargs):
        self.enum = kwargs.pop("enum", None)
        super(EnumField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        if isinstance(data, int):
            try:
                data = self.enum(data)
            except ValueError:
                self.fail("invalid")
        else:
            try:
                data = getattr(self.enum, data)
            except AttributeError:
                self.fail("invalid")
        return data.value

    def to_representation(self, value):
        return self.enum(value).name


class IsMailingListPublicOrIsMember(permissions.BasePermission):
    """
    Custom permission to only allow access to public lists or to lists the
    authenticated user is a member of.
    """

    def _get_mlist(self, obj):
        if isinstance(obj, MailingList):
            return obj
        if hasattr(obj, "mailinglist"):
            return obj.mailinglist

    def has_object_permission(self, request, view, obj):
        mlist = self._get_mlist(obj)
        if mlist is None:
            # This is not a object linked to a mailing-list.
            return True
        return is_mlist_authorized(request, mlist)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-03-09 00:56
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0015_mailinglist_id_3'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mailinglist',
            name='archive_policy',
            field=models.IntegerField(
                choices=[(0, 'never'), (1, 'private'), (2, 'public')],
                default=2),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#
import logging
from collections import namedtuple

from django.conf import settings
from django.db import models
from django.utils.timezone import now, utc

from hyperkitty.lib.analysis import compute_thread_order_and_depth
from .common import ModelCachedValue, VotesCachedValue


logger = logging.getLogger(__name__)


class Thread(models.Model):
    """
    A thread of archived email, from a mailing-list. It is identified by both
    the list name and the thread id.
    """
    mailinglist = models.ForeignKey(
        # Delete the model if the MailingList is deleted.
        "MailingList", related_name="threads", on_delete=models.CASCADE)
    thread_id = models.CharField(max_length=255, db_index=True)
    date_active = models.DateTimeField(db_index=True, default=now)
    category = models.ForeignKey(
        "ThreadCategory", related_name="threads", null=True,
        on_delete=models.SET_NULL)
    starting_email = models.OneToOneField(
        "Email", related_name="started_thread", null=True,
        on_delete=models.SET_NULL)

    def __init__(self, *args, **kwargs):
        super(Thread, self).__init__(*args, **kwargs)
        self.cached_values = {
            "participants_count": ParticipantsCount(self),
            "emails_count": EmailsCount(self),
            "subject": Subject(self),
            "votes": VotesCachedValue(self),
            "votes_total": VotesTotal(self),
        }

    class Meta:
        unique_together = ("mailinglist", "thread_id")

    def __str__(self):
        return self.subject

    @property
    def participants(self):
        """Set of email senders in this thread"""
        from .email import Email
        Participant = namedtuple("Participant", ["name", "address"])
        return [
            Participant(name=e["sender_name"], address=e["sender__address"])
            for e in Email.objects.filter(thread_id=self.id).values(
                "sender__address", "sender_name").distinct()
            ]

    @property
    def participants_count(self):
        return self.cached_values["participants_count"]()

    def replies_after(self, date):
        return self.emails.filter(date__gt=date)

    # def _get_category(self):
    #     if not self.category_id:
    #         return None
    #     return self.category_obj.name
    # def _set_category(self, name):
    #     if not name:
    #         self.category_id = None
    #         return
    #     session = object_session(self)
    #     try:
    #         category = session.query(Category).filter_by(name=name).one()
    #     except NoResultFound:
    #         category = Category(name=name)
    #         session.add(category)
    #     self.category_id = category.id
    # category = property(_get_category, _set_category)

    @property
    def emails_count(self):
        return self.cached_values["emails_count"]()

    @property
    def subject(self):
        return self.cached_values["subject"]()

    def get_votes(self):
        return self.cached_values["votes"]()

    @property
    def votes_total(self):
        return self.cached_values["votes_total"]()

    @property
    def prev_thread(self):  # TODO: Make it a relationship
        return Thread.objects.filter(
                mailinglist__id=self.mailinglist_id,
                date_active__lt=self.date_active
            ).order_by("-date_active").first()

    @property
    def next_thread(self):  # TODO: Make it a relationship
        return Thread.objects.filter(
                mailinglist__id=self.mailinglist_id,
                date_active__gt=self.date_active
            ).order_by("date_active").first()

    def is_unread_by(self, user):
        if not user.is_authenticated:
            return False
        try:
            last_view = LastView.objects.get(thread=self, user=user)
        except LastView.DoesNotExist:
            return True
        except LastView.MultipleObjectsReturned:
            last_view_duplicate, last_view = LastView.objects.filter(
                thread=self, user=user).order_by("view_date").all()
            last_view_duplicate.delete()
        return self.date_active.replace(tzinfo=utc) > last_view.view_date

    def find_starting_email(self):
        # Find and set the staring email if it was not specified
        from .email import Email  # circular import
        if self.starting_email is not None:
            return

        try:
            self.starting_email = self.emails.get(parent_id__isnull=True)
        except Email.DoesNotExist:
            self.starting_email = self.emails.order_by("date").first()
        except ValueError:
            # If the Thread is not created yet, the self.emails will raise a
            # ValueError exception. This happens at creation time because this
            # method is called by on_pre_save
            #
            # https://docs.djangoproject.com/en/4.1/releases/4.1/#reverse-foreign-key-changes-for-unsaved-model-instances
            return

    def on_pre_save(self):
        self.find_starting_email()

    def on_post_created(self):
        self.mailinglist.on_thread_added(self)

    def on_post_save(self):
        pass

    def on_post_delete(self):
        self.mailinglist.on_thread_deleted(self)

    def on_email_added(self, email):
        self.find_starting_email()
        self.date_active = email.date
        if self.starting_email is None:
            self.starting_email = email
        self.save()
        if not getattr(settings, "HYPERKITTY_BATCH_MODE", False):
            # Cache handling and thread positions will be handled at the end of
            # the import process.
            from hyperkitty.tasks import (
                compute_thread_positions, rebuild_thread_cache_new_email)
            rebuild_thread_cache_new_email(self.id)
            compute_thread_positions(self.id)

    def on_email_deleted(self, email):
        from hyperkitty.tasks import rebuild_thread_cache_new_email

        # update or cleanup thread
        if self.emails.count() == 0:
            self.delete()
        else:
            if self.starting_email is None:
                self.find_starting_email()
                self.save(update_fields=["starting_email"])
            compute_thread_order_and_depth(self)
            self.date_active = self.emails.order_by("-date").first().date
            rebuild_thread_cache_new_email(self.id)

    def on_vote_added(self, vote):
        from hyperkitty.tasks import rebuild_thread_cache_votes
        rebuild_thread_cache_votes(self.id)

    on_vote_deleted = on_vote_added


class ParticipantsCount(ModelCachedValue):

    cache_key = "participants_count"

    def get_value(self):
        return len(self.instance.participants)


class EmailsCount(ModelCachedValue):

    cache_key = "emails_count"

    def get_value(self):
        return self.instance.emails.count()


class Subject(ModelCachedValue):

    cache_key = "subject"

    def get_value(self):
        return self.instance.starting_email.subject


class VotesTotal(ModelCachedValue):

    cache_key = "votes_total"

    def get_value(self):
        votes = self.instance.get_votes()
        return votes["likes"] - votes["dislikes"]


class LastView(models.Model):
    thread = models.ForeignKey(
        "Thread", related_name="lastviews", on_delete=models.CASCADE)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="lastviews", on_delete=models.CASCADE)
    view_date = models.DateTimeField(auto_now=True)

    def __str__(self):
        """Unicode representation"""
        return "Last view of %s by %s was %s" % (
            str(self.thread), str(self.user),
            self.view_date.isoformat())

    def num_unread(self):
        if self.thread.date_active.replace(tzinfo=utc) <= self.view_date:
            return 0  # avoid the expensive query below
        else:
            return self.thread.emails.filter(date__gt=self.view_date).count()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2023 by the Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
# USA.

"""
Warm up the cache.

Author: Aurelien Bompard <abompard@fedoraproject.org>
"""

import datetime

from django.core.management.base import BaseCommand
from django.utils.timezone import now

from hyperkitty.management.utils import setup_logging
from hyperkitty.models import MailingList


class Command(BaseCommand):
    help = "Warm up the cache"

    def add_arguments(self, parser):
        parser.add_argument('mlists', nargs='*')
        parser.add_argument(
            '-m', '--months', type=int, default=1,
            help="number of months to cache")

    def handle(self, *args, **options):
        setup_logging(self, options["verbosity"])
        mlists = [
            MailingList.objects.get(name=name)
            for name in options["mlists"]
            ]
        if not mlists:
            mlists = MailingList.objects.order_by("name").all()
        for mlist in mlists:
            self.warm_up_mlist(mlist, options)

    def warm_up_mlist(self, mlist, options):
        if options["verbosity"] > 1:
            self.stdout.write("Warming up cache for %s" % mlist.name)
        # Recent data
        for cached_value in mlist.recent_cached_values:
            cached_value.warm_up()
        for thread in mlist.recent_threads:
            self.warm_up_thread(thread)
        # Other months
        month_start = now().replace(
            day=1, hour=0, minute=0, second=0, microsecond=0)
        for month_num in range(options["months"]):
            month_start = month_start - datetime.timedelta(days=1)
            month_start = month_start.replace(day=1)
            mlist.cached_values["participants_count_for_month"].warm_up(
                month_start.year, month_start.month)
            month_end = month_start + datetime.timedelta(days=32)
            month_end = month_end.replace(day=1)
            for thread in mlist.get_threads_between(month_start, month_end):
                self.warm_up_thread(thread)

    def warm_up_thread(self, thread):
        for cached_value in thread.cached_values.values():
            cached_value.warm_up()
        for email in thread.emails.all():
            for cached_value in email.cached_values.values():
                cached_value.warm_up()



# -*- coding: utf-8 -*-
# flake8: noqa

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hyperkitty', '0010_email_sender_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='email',
            name='parent',
            field=models.ForeignKey(
                related_name='children',
                on_delete=django.db.models.deletion.DO_NOTHING,
                blank=True, to='hyperkitty.Email', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from rest_framework import generics, serializers
from rest_framework.pagination import LimitOffsetPagination

from hyperkitty.models import ArchivePolicy, MailingList
from .utils import EnumField, IsMailingListPublicOrIsMember


class MailingListSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name='hk_api_mailinglist_detail', lookup_field="name",
        lookup_url_kwarg="mlist_fqdn")
    threads = serializers.HyperlinkedIdentityField(
        view_name='hk_api_thread_list', lookup_field="name",
        lookup_url_kwarg="mlist_fqdn")
    emails = serializers.HyperlinkedIdentityField(
        view_name='hk_api_email_list', lookup_field="name",
        lookup_url_kwarg="mlist_fqdn")
    archive_policy = EnumField(enum=ArchivePolicy)

    class Meta:
        model = MailingList
        fields = (
            "url", "name", "display_name", "description", "subject_prefix",
            "archive_policy", "created_at", "threads", "emails")
        lookup_field = "name"


class MailingListList(generics.ListAPIView):
    """List mailing-lists"""

    queryset = MailingList.objects.exclude(
        archive_policy=ArchivePolicy.private.value)
    ordering = ("name", )
    ordering_fields = ("name", "created_at")
    lookup_field = "name"
    serializer_class = MailingListSerializer
    pagination_class = LimitOffsetPagination


class MailingListDetail(generics.RetrieveAPIView):
    """Show a mailing-list"""

    queryset = MailingList.objects.all()
    lookup_field = "name"
    lookup_url_kwarg = "mlist_fqdn"
    serializer_class = MailingListSerializer
    permission_classes = [IsMailingListPublicOrIsMember]
    pagination_class = LimitOffsetPagination



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2023 by the Free Software Foundation, Inc.
#
# This file is part of HyperKitty.
#
# HyperKitty is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# HyperKitty is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# HyperKitty.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Aurelien Bompard <abompard@fedoraproject.org>
#

from contextlib import contextmanager

from django.conf import settings
from django.db.models.signals import (
    post_delete, post_init, post_save, pre_delete, pre_save)
from django.dispatch import receiver

from django_mailman3.signals import mailinglist_created, mailinglist_modified

from hyperkitty.lib.mailman import import_list_from_mailman
from hyperkitty.models.email import Attachment, Email
from hyperkitty.models.mailinglist import MailingList
from hyperkitty.models.profile import Profile
from hyperkitty.models.thread import Thread
from hyperkitty.models.vote import Vote


# Email

@receiver(post_init, sender=Email)
def Email_on_post_init(sender, **kwargs):
    kwargs["instance"].on_post_init()


@receiver(pre_save, sender=Email)
def Email_on_pre_save(sender, **kwargs):
    kwargs["instance"].on_pre_save()


@receiver(post_save, sender=Email)
def Email_on_post_save(sender, **kwargs):
    if kwargs["created"]:
        kwargs["instance"].on_post_created()
    else:
        kwargs["instance"].on_post_save()


@receiver(pre_delete, sender=Email)
def Email_on_pre_delete(sender, **kwargs):
    kwargs["instance"].on_pre_delete()


@receiver(post_delete, sender=Email)
def Email_on_post_delete(sender, **kwargs):
    kwargs["instance"].on_post_delete()


# Attachment

@receiver(pre_save, sender=Attachment)
def Attachment_on_pre_save(sender, **kwargs):
    kwargs["instance"].on_pre_save()


# MailingList

@receiver(pre_save, sender=MailingList)
def MailingList_set_list_id(sender, **kwargs):
    kwargs["instance"].on_pre_save()


# Profile

@receiver(post_save, sender=settings.AUTH_USER_MODEL)
def Profile_on_post_save(sender, **kwargs):
    user = kwargs["instance"]
    if not Profile.objects.filter(user=user).exists():
        Profile.objects.create(user=user)


# Thread

@receiver(pre_save, sender=Thread)
def Thread_on_pre_save(sender, **kwargs):
    kwargs["instance"].on_pre_save()


@receiver(post_save, sender=Thread)
def Thread_on_post_save(sender, **kwargs):
    if kwargs["created"]:
        kwargs["instance"].on_post_created()
    else:
        kwargs["instance"].on_post_save()


@receiver(post_delete, sender=Thread)
def Thread_on_post_delete(sender, **kwargs):
    kwargs["instance"].on_post_delete()


# Vote

@receiver(post_save, sender=Vote)
def Vote_on_post_save(sender, **kwargs):
    kwargs["instance"].on_post_save()


@receiver(post_delete, sender=Vote)
def Vote_on_post_delete(sender, **kwargs):
    kwargs["instance"].on_post_delete()


# Mailman signals

@receiver(mailinglist_created)
def on_mailinglist_created(sender, **kwargs):
    import_list_from_mailman(kwargs["list_id"])


@receiver(mailinglist_modified)
def on_mailinglist_modified(sender, **kwargs):
    import_list_from_mailman(kwargs["list_id"])


# Utility functions.

@contextmanager
def silenced_email_pre_delete():
    """Disable the handling of pre_delete signal for Email.

    This is required because when trying to delete the entire thread, it
    doesn't make sense to keep rebalancing the parent of the thread.  This is
    what the `on_pre_delete` signal handler does.

    We use this only when we are absolutely sure that we are going to delete
    the thread because we know it can land us a weird state where there are no
    parents or references to non-existent rows.
    """
    pre_delete.disconnect(Email_on_pre_delete, sender=Email)
    post_delete.disconnect(Email_on_post_delete, sender=Email)
    yield
    pre_delete.connect(Email_on_pre_delete, sender=Email)
    post_delete.connect(Email_on_post_delete, sender=Email)
